﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Diagnostics;
using System.Reflection;
using FDK;

namespace SSTFEditor
{
	// ※デザイナでは開けないので注意すること。

	partial class Cメインフォーム : Form
	{
		public readonly int nRevision = FileVersionInfo.GetVersionInfo( Assembly.GetExecutingAssembly().Location ).FilePrivatePart;
		public readonly int nBuild = FileVersionInfo.GetVersionInfo( Assembly.GetExecutingAssembly().Location ).FileBuildPart;

		public const int n最大音量 = 4;
		public const int n最小音量 = 1;

		public bool b未保存
		{
			get
			{
				return this._b未保存;
			}
			set
			{
				// まず値を保存。

				this._b未保存 = value;

				// ウィンドウのタイトルバーの文字列を修正。

				string str表示するファイル名 =
						( string.IsNullOrEmpty( this.str編集中のファイル名 ) ) ? Properties.Resources.NEW_FILENAME : this.str編集中のファイル名;

				if( this._b未保存 )
				{
					// 変更ありかつ未保存なら「*」を付ける

					this.Text = string.Format( "SSTFEditor rev.{1} (build.{2}) *[{0}]", str表示するファイル名, this.nRevision, this.nBuild );
					this.toolStripMenuItem上書き保存.Enabled = true;
					this.toolStripButton上書き保存.Enabled = true;
				}
				else
				{
					// 保存後変更がないなら「*」なない

					this.Text = string.Format( "SSTFEditor rev.{1} (build.{2}) [{0}]", str表示するファイル名, this.nRevision, this.nBuild );
					this.toolStripMenuItem上書き保存.Enabled = false;
					this.toolStripButton上書き保存.Enabled = false;
				}
			}
		}
		public bool b選択モードである
		{
			get { return ( this.toolStripButton選択モード.CheckState == CheckState.Checked ); }
		}
		public bool b編集モードである
		{
			get { return ( this.toolStripButton編集モード.CheckState == CheckState.Checked ); }
		}

		public int nGRID_PER_PART
		{
			get;
			protected set;
		}
		public int nGRID_PER_PIXEL
		{
			get { return this._nGRID_PER_PIXEL / ( this.toolStripComboBox譜面拡大率.SelectedIndex + 1 ); }
		}

		public bool b初期化完了 = false;
		public Config Config;
		public C選択モード 選択モード;
		public C編集モード 編集モード;
		public C譜面 譜面;
		public CUndoRedo UndoRedo;
		public Cクリップボード クリップボード;
		public Size sz譜面パネル
		{
			get { return this.pictureBox譜面パネル.ClientSize; }
		}
		public StrokeStyleT.Eチップ e現在のチップ種別
		{
			get
			{
				return this._e現在のチップ種別; 
			}
			set
			{
				this._e現在のチップ種別 = value;
				this.label現在のチップ種別.Text = this.dicチップ種別名[ value ];
			}
		}
		public int n現在のチップ音量 = Cメインフォーム.n最大音量;

		public Cメインフォーム()
		{
			InitializeComponent();

			this.tActアプリの起動();
		}

		public void t選択モードに切替えて関連GUIを設定する()
		{
			this.toolStripButton選択モード.CheckState = CheckState.Checked;
			this.toolStripButton編集モード.CheckState = CheckState.Unchecked;

			this.toolStripMenuItem選択モード.CheckState = CheckState.Checked;
			this.toolStripMenuItem編集モード.CheckState = CheckState.Unchecked;

			this.label現在のチップ種別.Text = "----";

			this.t譜面をリフレッシュする();
		}
		public void t編集モードに切替えて関連GUIを設定する()
		{
			this.選択モード.t全チップの選択を解除する();
			this.t譜面をリフレッシュする();

			this.toolStripButton選択モード.CheckState = CheckState.Unchecked;
			this.toolStripButton編集モード.CheckState = CheckState.Checked;

			this.toolStripMenuItem選択モード.CheckState = CheckState.Unchecked;
			this.toolStripMenuItem編集モード.CheckState = CheckState.Checked;
		}
		public void t選択チップの有無に応じて編集用GUIのEnabledを設定する()
		{
			bool b譜面上に選択チップがある = this.b選択チップが１個以上ある;
			bool bクリップボードに選択チップがある = ( this.クリップボード != null ) && ( this.クリップボード.nセル数 > 0 );


			// 編集メニュー

			this.toolStripMenuItemコピー.Enabled = b譜面上に選択チップがある;
			this.toolStripMenuItem切り取り.Enabled = b譜面上に選択チップがある;
			this.toolStripMenuItem貼り付け.Enabled = bクリップボードに選択チップがある;
			this.toolStripMenuItem削除.Enabled = b譜面上に選択チップがある;


			// ツールバー

			this.toolStripButtonコピー.Enabled = b譜面上に選択チップがある;
			this.toolStripButton切り取り.Enabled = b譜面上に選択チップがある;
			this.toolStripButton貼り付け.Enabled = bクリップボードに選択チップがある;
			this.toolStripButton削除.Enabled = b譜面上に選択チップがある;


			// 右メニュー

			this.toolStripMenuItem選択チップのコピー.Enabled = b譜面上に選択チップがある;
			this.toolStripMenuItem選択チップの切り取り.Enabled = b譜面上に選択チップがある;
			this.toolStripMenuItem選択チップの貼り付け.Enabled = bクリップボードに選択チップがある;
			this.toolStripMenuItem選択チップの削除.Enabled = b譜面上に選択チップがある;
		}
		public void t譜面をリフレッシュする()
		{
			this.pictureBox譜面パネル.Refresh();
		}
		public void tUndoRedo用GUIのEnabledを設定する()
		{
			bool bUndoの可否 = ( this.UndoRedo.nUndo可能な回数 > 0 );
			bool bRedoの可否 = ( this.UndoRedo.nRedo可能な回数 > 0 );

			this.toolStripMenuItem元に戻す.Enabled = bUndoの可否;
			this.toolStripMenuItemやり直す.Enabled = bRedoの可否;
			this.toolStripButton元に戻す.Enabled = bUndoの可否;
			this.toolStripButtonやり直す.Enabled = bRedoの可否;
		}
		public void t選択モードのコンテクストメニューを表示する( int x, int y )
		{
			// メニューを表示。

			this.contextMenuStrip譜面右メニュー.Show( this.pictureBox譜面パネル, x, y );

	
			// メニューを表示した時のマウス座標を控えておく。

			this.pt選択モードのコンテクストメニューを開いたときのマウスの位置 = new Point( x, y );
		}
		public void t譜面を縦スクロールする( int nスクロール量grid )
		{
			int n現在の位置grid = this.vScrollBar譜面用垂直スクロールバー.Value;
			int nスクロール後の位置grid = this.vScrollBar譜面用垂直スクロールバー.Value + nスクロール量grid;
			int n最小値grid = this.vScrollBar譜面用垂直スクロールバー.Minimum;
			int n最大値grid = ( this.vScrollBar譜面用垂直スクロールバー.Maximum + 1 ) - this.vScrollBar譜面用垂直スクロールバー.LargeChange;
			
			if( nスクロール後の位置grid < n最小値grid )
			{
				nスクロール後の位置grid = n最小値grid;
			}
			else if( nスクロール後の位置grid > n最大値grid )
			{
				nスクロール後の位置grid = n最大値grid;
			}

			this.vScrollBar譜面用垂直スクロールバー.Value = nスクロール後の位置grid;
		}

		#region [ protected ]
		//-----------------

		protected readonly Dictionary<StrokeStyleT.Eチップ, string> dicチップ種別名
			#region [ **** ]
			//-----------------
			= new Dictionary<StrokeStyleT.Eチップ, string>() {
				{ StrokeStyleT.Eチップ.Bass, "BassDrum" },
				{ StrokeStyleT.Eチップ.BPM, "BPM" },
				{ StrokeStyleT.Eチップ.China, "China" },
				{ StrokeStyleT.Eチップ.HiHat_Close, "HiHat(Close)" },
				{ StrokeStyleT.Eチップ.HiHat_Foot, "FootPedal" },
				{ StrokeStyleT.Eチップ.HiHat_HalfOpen, "HiHat(HarfOpen)" },
				{ StrokeStyleT.Eチップ.HiHat_Open, "HiHat(Open)" },
				{ StrokeStyleT.Eチップ.LeftCrash, "Crash" },
				{ StrokeStyleT.Eチップ.Ride, "Ride" },
				{ StrokeStyleT.Eチップ.Ride_Cup, "Ride(Cup)" },
				{ StrokeStyleT.Eチップ.RightCrash, "Crash" },
				{ StrokeStyleT.Eチップ.Snare, "Snare" },
				{ StrokeStyleT.Eチップ.Snare_ClosedRim, "Snare(CloseRimShot)" },
				{ StrokeStyleT.Eチップ.Snare_Ghost, "Snare(Ghost)" },
				{ StrokeStyleT.Eチップ.Snare_OpenRim, "Snare(OpenRimShot)" },
				{ StrokeStyleT.Eチップ.Splash, "Splash" },
				{ StrokeStyleT.Eチップ.Tom1, "HighTom" },
				{ StrokeStyleT.Eチップ.Tom1_Rim, "HighTom(RimShot)" },
				{ StrokeStyleT.Eチップ.Tom2, "LowTom" },
				{ StrokeStyleT.Eチップ.Tom2_Rim, "LowTom(RimShot)" },
				{ StrokeStyleT.Eチップ.Tom3, "FloorTom" },
				{ StrokeStyleT.Eチップ.Tom3_Rim, "FloorTom(RimShot)" },
				{ StrokeStyleT.Eチップ.Unknown, "" },
				{ StrokeStyleT.Eチップ.小節線, "" },
				{ StrokeStyleT.Eチップ.背景動画, "" },
				{ StrokeStyleT.Eチップ.拍線, "" },
				{ StrokeStyleT.Eチップ.小節メモ, "" },
			};
			//-----------------
			#endregion
	
		protected bool _b未保存 = false;
		protected StrokeStyleT.Eチップ _e現在のチップ種別;
		protected bool b選択チップが１個以上ある
		{
			get
			{
				if( this.譜面.Score != null &&
					this.譜面.Score.listチップ != null &&
					this.譜面.Score.listチップ.Count > 0 )
				{
					foreach( var chip in this.譜面.Score.listチップ )
					{
						if( chip.b選択が確定している )
							return true;
					}
				}
				return false;
			}
		}
		protected int n現在のガイド間隔 = 0;
		protected int _nGRID_PER_PIXEL = 1;
		protected Point pt選択モードのコンテクストメニューを開いたときのマウスの位置;

		/// <summary>
		/// <para>システムフォルダ。SSTFEditor.exe と StrokeStyleT.exe が格納されているフォルダへのパス。</para>
		/// <para>（例："c:\Program Files\StrokeStyleT\" ）</para>
		/// </summary>
		protected string strシステムフォルダ = "";
		/// <summary>
		/// <para>ユーザのアプリデータフォルダ。（例: "C:\Users\(User名)\ApplicationData\SSTFEditor\"）</para>
		/// </summary>
		protected string strユーザフォルダ = "";
		/// <summary>
		/// <para>作業フォルダ。ダイアログでカレントフォルダを移動したりすると作業フォルダも移動する。</para>
		/// </summary>
		protected string str作業フォルダ = "";
		/// <summary>
		/// <para>ファイル名のみ。（例："test.sstf"）</para>
		/// </summary>
		protected string str編集中のファイル名 = "";
		/// <summary>
		/// <para>システムフォルダ＋プレイヤー名。（例："C:\Program Files\StrokeStyleT\StrokeStyleT.exe"）</para>
		/// </summary>
		protected string strプレイヤーの絶対パスファイル名 = "";
		protected string str最後にプレイヤーに渡した一時ファイル名 = "";
	

		// アクションメソッド

		protected void tActアプリの起動()
		{
			this.nGRID_PER_PART = int.Parse( Properties.Resources.GRID_PER_PART );
			this._nGRID_PER_PIXEL = int.Parse( Properties.Resources.GRID_PER_PIXEL );

			#region [ システムフォルダ、ユーザフォルダ、作業フォルダ、プレイヤーの絶対パスファイル名 を取得する。]
			//-----------------
#if DEBUG
			this.strシステムフォルダ = Environment.CurrentDirectory;		// DEBUG 時に限り、システムフォルダはカレントフォルダ（VC# で設定した作業フォルダ）になる。
#else
			this.strシステムフォルダ = Path.GetDirectoryName( Application.ExecutablePath );	// DEBUG 以外は、exe の存在するフォルダ。
#endif
			this.strユーザフォルダ = Path.Combine( Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData ), @"SSTFEditor" );
			if( !Directory.Exists( this.strユーザフォルダ ) )					// なければ作成する。
				Directory.CreateDirectory( this.strユーザフォルダ );

			this.str作業フォルダ = Environment.GetFolderPath( Environment.SpecialFolder.MyDocuments ); // 作業フォルダの初期値としてユーザのマイドキュメントフォルダを設定する。

			this.strプレイヤーの絶対パスファイル名 = Path.Combine( this.strシステムフォルダ, Properties.Resources.PLAYER_NAME );
			//-----------------
			#endregion
			

			// Config.xml を読み込む。

			this.Config = Config.t読み込む( Path.Combine( this.strユーザフォルダ, Properties.Resources.CONFIG_FILE_NAME ) );
			

			// デザイナでは追加できないイベントを手動で追加する。

			this.splitContainer分割パネルコンテナ.MouseWheel += new MouseEventHandler( splitContainer分割パネルコンテナ_MouseWheel );

			
			// 最近使ったファイル一覧を更新する。

			this.tConfigのRecentUsedFilesをFileメニューへ追加する();

			
			// その他の初期化。

			this.tAct新規作成();
			this.tActガイド間隔を変更する( 16 );	// 初期は 1/16 間隔。
			this.tAct譜面拡大率を変更する( 1 );		// 初期は 標準。


			// 完了。

			this.b初期化完了 = true;
		}
		protected void tActアプリの終了()
		{
			if( File.Exists( this.str最後にプレイヤーに渡した一時ファイル名 ) )	// 一時ファイルが残っていれば削除。
				File.Delete( this.str最後にプレイヤーに渡した一時ファイル名 );

			this.Config.t保存する( Path.Combine( this.strユーザフォルダ, Properties.Resources.CONFIG_FILE_NAME ) );
			
			FDK.C共通.tDisposeする( ref this.譜面 );
			FDK.C共通.tDisposeする( ref this.選択モード );
		}
		protected void tAct新規作成()
		{
			#region [ 未保存なら保存する。→ キャンセルされた場合はここで中断。]
			//-----------------
			if( this.t未保存なら保存する() == DialogResult.Cancel )
				return;	// 中断
			//-----------------
			#endregion

			this.tエディタを初期化する();
		}
		protected void tAct開く()
		{
			#region [ 未保存なら保存する。→ キャンセルされた場合はここで中断。]
			//-----------------
			if( this.t未保存なら保存する() == DialogResult.Cancel )
				return;	// 中断
			//-----------------
			#endregion

			#region [ 「ファイルを開く」ダイアログでファイルを選択する。 ]
			//-----------------
			var dialog = 
				new OpenFileDialog() {
					Title = Properties.Resources.MSG_ファイル選択ダイアログのタイトル,
					Filter = Properties.Resources.MSG_ファイル選択ダイアログのフィルタ,
					FilterIndex = 1,
					InitialDirectory = this.str作業フォルダ,
				};
			var result = dialog.ShowDialog( this );

			// メインフォームを再描画してダイアログを完全に消す。
			this.Refresh();

			if( result != DialogResult.OK )
				return;		// OKじゃないならここまで。
			//-----------------
			#endregion

			this.tファイルを読み込む( dialog.FileName );
		}
		protected void tAct指定されたファイルを開く( string strFileNames )
		{
			#region [ 未保存なら保存する。→ キャンセルされた場合はここで中断。]
			//-----------------
			if( this.t未保存なら保存する() == DialogResult.Cancel )
				return;	// 中断
			//-----------------
			#endregion

			this.tファイルを読み込む( strFileNames );
		}
		protected void tAct上書き保存()
		{
			#region [ ファイル名がない → 初めての保存と見なし、ファイル保存ダイアログで保存ファイル名を指定させる。 ]
			//-----------------
			if( string.IsNullOrEmpty( this.str編集中のファイル名 ) )
			{
				// ダイアログでファイル名を取得する。

				string str絶対パスファイル名 = this.tファイル保存ダイアログを開いてファイル名を取得する();

				if( string.IsNullOrEmpty( str絶対パスファイル名 ) )
					return;	// ファイル保存ダイアログがキャンセルされたのならここで打ち切り。

				this.str作業フォルダ = Path.GetDirectoryName( str絶対パスファイル名 );	// ダイアログでディレクトリを変更した場合、カレントディレクトリも変更されている。
				this.str編集中のファイル名 = Path.GetFileName( str絶対パスファイル名 );
			}
			//-----------------
			#endregion

			this.tAct上書き保存( Path.Combine( this.str作業フォルダ, this.str編集中のファイル名 ), false );
		}
		protected void tAct上書き保存( string strファイルの絶対パス, bool bプレイヤー用一時ファイルである )
		{
			#region [「保存中です」ポップアップを表示する。 ]
			//-----------------
			var msg = new CPopupメッセージ( Properties.Resources.MSG_保存中です + Environment.NewLine + Properties.Resources.MSG_しばらくお待ち下さい );
			msg.Owner = this;
			msg.Show();
			msg.Refresh();
			//-----------------
			#endregion
			#region [ 選択モードだったなら選択を解除する。]
			//-----------------
			if( this.b選択モードである )
				this.選択モード.t全チップの選択を解除する();
			//-----------------
			#endregion
			#region [ 出力するパス内の背景動画を検索する。]
			//-----------------
			譜面.Score.str背景動画ファイル名 = C共通.t指定した拡張子を持つファイルを検索し最初に見つけたファイルの絶対パスを返す( Path.GetDirectoryName( strファイルの絶対パス ), StrokeStyleT.Config.listデフォルト拡張子 );
			this.textBox背景動画.Text = Path.GetFileName( 譜面.Score.str背景動画ファイル名 );
			//-----------------
			#endregion
			#region [ SSTFファイルを出力する。]
			//-----------------
			this.譜面.tSSTFファイルを書き出す( strファイルの絶対パス );
			//-----------------
			#endregion
			#region [ 出力したファイルのパスを、[ファイル]メニューの最近使ったファイル一覧に追加する。 ]
			//-----------------
			if( !bプレイヤー用一時ファイルである )
			{
				this.Config.tファイルを最近使ったファイルの一覧に追加する( strファイルの絶対パス );
				this.tConfigのRecentUsedFilesをFileメニューへ追加する();
			}
			//-----------------
			#endregion

			#region [「保存中です」ポップアップを閉じる。]
			//-----------------
			msg.Close();
			//-----------------
			#endregion


			// 最後に、ダイアログのゴミなどを消すために再描画。

			this.Refresh();

			if( !bプレイヤー用一時ファイルである )
				this.b未保存 = false;
		}
		protected void tAct名前を付けて保存()
		{
			#region [ ユーザに保存ファイル名を入力させる。]
			//-----------------

			// ファイル保存ダイアログを表示し、出力するファイル名を指定させる。キャンセルされたらここで中断。

			string str絶対パスファイル名 = this.tファイル保存ダイアログを開いてファイル名を取得する();
			
			if( string.IsNullOrEmpty( str絶対パスファイル名 ) )
				return;	// 中断


			// フォルダ名とファイル名を更新。

			this.str作業フォルダ = Path.GetDirectoryName( str絶対パスファイル名 );
			this.str編集中のファイル名 = Path.GetFileName( str絶対パスファイル名 );
			//-----------------
			#endregion

			this.tAct上書き保存();

			this.b未保存 = true;	// ウィンドウタイトルに表示されているファイル名を変更するため一度 true にする。
			this.b未保存 = false;
		}
		protected void tAct終了()
		{
			this.Close();
		}
		protected void tAct元に戻す()
		{
			// Undo する対象を UndoRedoリストから取得する。

			var cell = this.UndoRedo.tUndoするセルを取得して返す();
			if( cell == null )
				return;		// なければ中断


			// Undo を実行する。

			cell.tUndoを実行する();


			// GUIを再描画する。

			this.tUndoRedo用GUIのEnabledを設定する();
			this.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tActやり直す()
		{
			// Redo する対象を UndoRedoリストから取得する。

			var cell = this.UndoRedo.tRedoするセルを取得して返す();
			if( cell == null )
				return;	// なければ中断


			// Redo を実行する。
	
			cell.tRedoを実行する();


			// GUI を再描画する。

			this.tUndoRedo用GUIのEnabledを設定する();
			this.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tAct切り取り()
		{
			// 譜面にフォーカスが来てないなら何もしない。

			if( !this.pictureBox譜面パネル.Focused )
				return;


			// 切り取り ＝ コピー ＋ 削除

			this.tActコピー();
			this.tAct削除();
		}
		protected void tActコピー()
		{
			// 譜面にフォーカスが来てないなら何もしない。

			if( !this.pictureBox譜面パネル.Focused )
				return;


			// コピーする。

			this.クリップボード.t現在選択されているチップをボードにコピーする();


			// 画面を再描画する。

			this.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tAct貼り付け( int n貼り付けを開始する譜面内絶対位置grid )
		{
			#region [ 譜面にフォーカスが来てないなら何もしない。 ]
			//-----------------
			if( !this.pictureBox譜面パネル.Focused )
				return;
			//-----------------
			#endregion

			this.クリップボード.tチップを指定位置から貼り付ける( n貼り付けを開始する譜面内絶対位置grid );
		}
		protected void tAct削除()
		{
			#region [ 譜面にフォーカスが来てないなら何もしない。 ]
			//-----------------
			if( !this.pictureBox譜面パネル.Focused )
				return;
			//-----------------
			#endregion

			this.UndoRedo.tトランザクション記録を開始する();


			// 譜面が持つすべてのチップについて、選択されているチップがあれば削除する。
			
			for( int i = this.譜面.Score.listチップ.Count - 1; i >= 0; i-- )
			{
				var chip = this.譜面.Score.listチップ[ i ];

				if( chip.b選択が確定している )
				{
					var chip変更前 = new StrokeStyleT.Cチップ( chip );

					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							変更対象.CopyFrom( 変更前 );
							this.譜面.Score.listチップ.Add( 変更対象 );
							this.譜面.Score.listチップ.Sort();
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.譜面.Score.listチップ.Remove( 変更対象 );
							this.b未保存 = true;
						},
						chip, chip変更前, null, null, null );

					this.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}

			this.UndoRedo.tトランザクション記録を終了する();

			this.tUndoRedo用GUIのEnabledを設定する();
			this.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tActすべて選択()
		{
			// 編集モードなら強制的に選択モードにする。

			if( this.b編集モードである )
				this.t選択モードに切替えて関連GUIを設定する();


			// 全チップを選択する。

			this.選択モード.t全チップを選択する();
		}
		protected void tAct選択モードにする()
		{
			this.t選択モードに切替えて関連GUIを設定する();
		}
		protected void tAct編集モードにする()
		{
			this.t編集モードに切替えて関連GUIを設定する();
		}
		protected void tActモードを切替える()
		{
			if( this.b選択モードである )
			{
				this.t編集モードに切替えて関連GUIを設定する();
			}
			else
			{
				this.t選択モードに切替えて関連GUIを設定する();
			}
		}
		protected void tAct検索()
		{
			this.選択モード.t検索する();			// モードによらず、検索はすべて選択モードが行う。	
		}
		protected void tActガイド間隔を変更する( int n分 )
		{
			#region [ 引数チェック。]
			//-----------------
			if( !new List<int>() { 4, 6, 8, 12, 16, 24, 32, 48, 64, 128, 0 }.Contains( n分 ) )
				throw new ArgumentException( string.Format( "不正なガイド間隔({0})が指定されました。", n分 ) );
			//-----------------
			#endregion

			#region [ 新しいガイド間隔を設定する。]
			//-----------------
			this.n現在のガイド間隔 = n分;
			this.譜面.t現在のガイド間隔を変更する( n分 );	// 譜面にも伝達。
			//-----------------
			#endregion

			#region [ ガイド間隔関連GUI（メニュー、コンボボックス）を更新する。]
			//-----------------

			// 一度すべてのガイド間隔メニューのチェックをはずし、制定された分数のメニューのみチェックする。 
			
			this.toolStripMenuItemガイド間隔4分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔6分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔8分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔12分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔16分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔24分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔32分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔48分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔64分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔128分.CheckState = CheckState.Unchecked;
			this.toolStripMenuItemガイド間隔フリー.CheckState = CheckState.Unchecked;

			switch( n分 )
			{
				// Menu と ComboBox の２つを変更することでイベントが２つ発生し、最終的に
				// tActガイド間隔を変更する() を立て続けに２回呼び出してしまうことになるが……、まぁよしとする。（汗

				case 4:
					this.toolStripMenuItemガイド間隔4分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 0;
					break;

				case 6:
					this.toolStripMenuItemガイド間隔6分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 1;
					break;

				case 8:
					this.toolStripMenuItemガイド間隔8分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 2;
					break;

				case 12:
					this.toolStripMenuItemガイド間隔12分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 3;
					break;

				case 16:
					this.toolStripMenuItemガイド間隔16分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 4;
					break;

				case 24:
					this.toolStripMenuItemガイド間隔24分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 5;
					break;

				case 32:
					this.toolStripMenuItemガイド間隔32分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 6;
					break;

				case 48:
					this.toolStripMenuItemガイド間隔48分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 7;
					break;

				case 64:
					this.toolStripMenuItemガイド間隔64分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 8;
					break;

				case 128:
					this.toolStripMenuItemガイド間隔128分.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 9;
					break;

				case 0:
					this.toolStripMenuItemガイド間隔フリー.CheckState = CheckState.Checked;
					this.toolStripComboBoxガイド間隔.SelectedIndex = 10;
					break;
			}
			//-----------------
			#endregion

			#region [ 画面を再描画する。]
			//-----------------
			this.pictureBox譜面パネル.Invalidate();
			//-----------------
			#endregion
		}
		protected void tActガイド間隔を拡大する()
		{
			switch( this.n現在のガイド間隔 )
			{
				case 4: break;
				case 6: this.tActガイド間隔を変更する( 4 ); break;
				case 8: this.tActガイド間隔を変更する( 6 ); break;
				case 12: this.tActガイド間隔を変更する( 8 ); break;
				case 16: this.tActガイド間隔を変更する( 12 ); break;
				case 24: this.tActガイド間隔を変更する( 16 ); break;
				case 32: this.tActガイド間隔を変更する( 24 ); break;
				case 48: this.tActガイド間隔を変更する( 32 ); break;
				case 64: this.tActガイド間隔を変更する( 48 ); break;
				case 128: this.tActガイド間隔を変更する( 64 ); break;
				case 0: this.tActガイド間隔を変更する( 128 ); break;
			}
		}
		protected void tActガイド間隔を縮小する()
		{
			switch( this.n現在のガイド間隔 )
			{
				case 4: this.tActガイド間隔を変更する( 6 ); break;
				case 6: this.tActガイド間隔を変更する( 8 ); break;
				case 8: this.tActガイド間隔を変更する( 12 ); break;
				case 12: this.tActガイド間隔を変更する( 16 ); break;
				case 16: this.tActガイド間隔を変更する( 24 ); break;
				case 24: this.tActガイド間隔を変更する( 32 ); break;
				case 32: this.tActガイド間隔を変更する( 48 ); break;
				case 48: this.tActガイド間隔を変更する( 64 ); break;
				case 64: this.tActガイド間隔を変更する( 128 ); break;
				case 128: this.tActガイド間隔を変更する( 0 ); break;
				case 0: break;
			}
		}
		protected void tAct譜面拡大率を変更する( int n倍 )
		{
			#region [ 引数チェック。]
			//-----------------
			if( n倍 < 1 || n倍 > 10 )
				throw new ArgumentException( string.Format( "不正な譜面拡大率({0})が指定されました。", n倍 ) );
			//-----------------
			#endregion

			this.toolStripComboBox譜面拡大率.SelectedIndex = n倍 - 1;

			this.t譜面をリフレッシュする();
		}
		protected void tAct最初から再生する()
		{
			this.tAct指定された小節の先頭から再生する( 0 );
		}
		protected void tAct現在位置から再生する()
		{
			int n位置grid;
			int n小節番号 = this.譜面.t譜面内絶対位置gridにおける小節の番号と小節先頭の位置gridを返す( this.譜面.n現在のカレントラインの譜面内絶対位置grid, out n位置grid );

			this.tAct指定された小節の先頭から再生する( n小節番号 );
		}
		protected void tAct現在位置からBGMのみ再生する()
		{
			int n位置grid;
			int n小節番号 = this.譜面.t譜面内絶対位置gridにおける小節の番号と小節先頭の位置gridを返す( this.譜面.n現在のカレントラインの譜面内絶対位置grid, out n位置grid );

			this.tAct指定された小節の先頭から再生する( n小節番号, false );
		}
		protected void tAct指定された小節の先頭から再生する( int n小節番号 )
		{
			this.tAct指定された小節の先頭から再生する( n小節番号, true );
		}
		protected void tAct指定された小節の先頭から再生する( int n小節番号, bool b仮想ドラム使用 )
		{
			if( string.IsNullOrEmpty( this.strプレイヤーの絶対パスファイル名 ) || !File.Exists( this.strプレイヤーの絶対パスファイル名 ) )
				return;


			// 前回のテンポラリファイルが存在していれば削除する。

			if( File.Exists( this.str最後にプレイヤーに渡した一時ファイル名 ) )
				File.Delete( this.str最後にプレイヤーに渡した一時ファイル名 );


			// 譜面を新しくテンポラリファイルとして出力する。

			while( true )
			{
				this.str最後にプレイヤーに渡した一時ファイル名 = Path.Combine( this.str作業フォルダ, Path.GetRandomFileName() );

				if( File.Exists( this.str最後にプレイヤーに渡した一時ファイル名 ) )
					continue;	// （まずないだろうが）同名ファイルが存在していれば、別の名前にする。

				break;
			}

			this.tAct上書き保存( this.str最後にプレイヤーに渡した一時ファイル名, true );	// 一時ファイルなので、「最近使ったファイル一覧」には残さない。


			// プレイヤーを起動する。

			Process.Start( this.strプレイヤーの絶対パスファイル名, "\"" + this.str最後にプレイヤーに渡した一時ファイル名 + "\" -p " + n小節番号.ToString() + (( b仮想ドラム使用 ) ? @" -h" : @"") );
		}
		protected void tAct再生を停止する()
		{
			if( string.IsNullOrEmpty( this.strプレイヤーの絶対パスファイル名 ) || !File.Exists( this.strプレイヤーの絶対パスファイル名 ) )
				return;

			Process.Start( this.strプレイヤーの絶対パスファイル名, "-s" );
		}
		protected void tActオプションを設定する()
		{
			var dialog = new Cオプションダイアログ();

			// Config の現在の値を反映。

			dialog.checkBoxオートフォーカス.CheckState = ( this.Config.AutoFocus ) ? CheckState.Checked : CheckState.Unchecked;
			dialog.checkBox最近使用したファイル.CheckState = ( this.Config.ShowRecentUsedFiles ) ? CheckState.Checked : CheckState.Unchecked;
			dialog.numericUpDown最近使用したファイルの最大表示個数.Value = this.Config.MaxOfUsedRecentFiles;


			// 決定された値を Config に反映。

			if( dialog.ShowDialog( this ) == DialogResult.OK )
			{
				this.Config.AutoFocus = dialog.checkBoxオートフォーカス.Checked;
				this.Config.ShowRecentUsedFiles = dialog.checkBox最近使用したファイル.Checked;
				this.Config.MaxOfUsedRecentFiles = (int) dialog.numericUpDown最近使用したファイルの最大表示個数.Value;


				// [ファイル] メニューを修正。

				this.tConfigのRecentUsedFilesをFileメニューへ追加する();
			}


			// 画面を再描画してダイアログのゴミを消す。

			this.Refresh();
		}
		protected void tActバージョンを表示する()
		{
			using( var dialog = new Cバージョン表示ダイアログ() )
				dialog.ShowDialog( this );
		}
		protected void tAct小節長倍率を変更する( int n小節番号 )
		{
			// 変更後倍率を取得する。

			double db変更後倍率 = 1.0;

			#region [ 小節長倍率をユーザに入力させる。入力がキャンセルされたらここで終了。]
			//-----------------

			// 現在の小節長倍率を取得する。

			var db現在の小節長倍率 = this.譜面.Score.t小節長倍率を取得する( n小節番号 );


			// ダイアログを表示し、小節長を取得する。

			var dialog = new C小節長倍率入力ダイアログ( n小節番号 ) {
				f倍率 = (float) db現在の小節長倍率,
				b後続変更 = false,
			};
			if( dialog.ShowDialog( this ) != DialogResult.OK )
				return;		// キャンセルされたらここで中断。

			//-----------------
			#endregion

			db変更後倍率 = (double) dialog.f倍率;


			// 後続もすべて変更する？

			int n変更開始小節番号 = n小節番号;
			int n変更終了小節番号 = ( dialog.b後続変更 ) ?
				this.譜面.Score.n最大小節番号 :		// する
				n小節番号;							// しない


			// 変更開始。

			this.UndoRedo.tトランザクション記録を開始する();

			for( int i = n変更開始小節番号; i <= n変更終了小節番号; i++ )
			{
				double db変更前倍率 = this.譜面.Score.t小節長倍率を取得する( i );

				#region [ 新しい小節長倍率の登録。]
				//-----------------
				var cell = new CUndoRedoセル<double>(
					null,
					( 変更対象, 変更前, 変更後, 小節番号, 任意2 ) => {	// Undo
						this.譜面.Score.t小節長倍率を設定する( (int) 小節番号, 変更前 );
						this.b未保存 = true;
					},
					( 変更対象, 変更前, 変更後, 小節番号, 任意2 ) => {	// Redo
						this.譜面.Score.t小節長倍率を設定する( (int) 小節番号, 変更後 );
						this.b未保存 = true;
					},
					0.0, db変更前倍率, db変更後倍率, i, null );

				this.UndoRedo.tセルを追加する( cell );
				cell.tRedoを実行する();
				//-----------------
				#endregion
				#region [ チップの移動または削除。]
				//-----------------
				int n変化量grid = (int) ( ( db変更後倍率 - db変更前倍率 ) * this.nGRID_PER_PART );

				for( int j = this.譜面.Score.listチップ.Count - 1; j >= 0; j-- )	// 削除する場合があるので後ろからカウント
				{
					var chip = this.譜面.Score.listチップ[ j ];

					// (A) 変更対象の小節内のチップ　→　移動なし。カウント変更あり。小節はみ出しチェックあり。

					if( chip.n小節番号 == i )
					{
						#region [ (a) 小節からはみ出したチップは、削除する。]
						//-----------------
						int n次小節の先頭位置grid = this.譜面.t小節先頭の譜面内絶対位置gridを返す( i ) + (int) ( db変更後倍率 * this.nGRID_PER_PART );

						if( chip.n譜面内絶対位置grid >= n次小節の先頭位置grid )
						{
							var chip変更前 = new StrokeStyleT.Cチップ( chip );

							var cc = new CUndoRedoセル<StrokeStyleT.Cチップ>(
								null,
								( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
									変更対象.CopyFrom( 変更前 );
									this.譜面.Score.listチップ.Add( 変更対象 );
									this.譜面.Score.listチップ.Sort();
									this.b未保存 = true;
								},
								( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
									this.譜面.Score.listチップ.Remove( 変更対象 );
									this.b未保存 = true;
								},
								chip, chip変更前, null, null, null );

							this.UndoRedo.tセルを追加する( cc );
							cc.tRedoを実行する();
						}
						//-----------------
						#endregion
						#region [ (b) 小節からはみ出さなかったチップは、カウント（小節解像度と小節内位置）を変更する。]
						//-----------------
						else
						{
							// 解像度が小さいと計算結果がおかしくなるので、無条件に解像度を上げる。

							int n充分に大きな値 = this.nGRID_PER_PART;	// 充分に大きい数なら何でもいい。

							if( chip.n小節解像度 < n充分に大きな値 )	// 何度も乗算されたら値が肥大化しまくるので、制限をかける。
							{
								chip.n小節解像度 *= n充分に大きな値;
								chip.n小節内位置 *= n充分に大きな値;
							}


							// 小節解像度を変更する。（小節内位置は変更なし。）

							chip.n小節解像度 = (int) ( db変更後倍率 * chip.n小節解像度 );
						}
						//-----------------
						#endregion
					}

					// (B) 変更対象より先の小節内のチップ　→　移動あり。カウントなし。小節はみ出しチェックなし。

					else if( chip.n小節番号 > i )
					{
						#region [ チップを n変化量grid 移動する。 ]
						//-----------------
						var cc = new CUndoRedoセル<StrokeStyleT.Cチップ>(
							null,
							( 変更対象, 変更前, 変更後, 変化量grid, 任意2 ) => {	// Undo
								 変更対象.n譜面内絶対位置grid -= (int) 変化量grid;
								 this.b未保存 = true;
							 },
							 ( 変更対象, 変更前, 変更後, 変化量grid, 任意2 ) => {	// Redo
								 変更対象.n譜面内絶対位置grid += (int) 変化量grid;
								 this.b未保存 = true;
							 },
							 chip, null, null, n変化量grid, null );

						this.UndoRedo.tセルを追加する( cc );
						cc.tRedoを実行する();
						//-----------------
						#endregion
					}
				}
				//-----------------
				#endregion
			}

			this.UndoRedo.tトランザクション記録を終了する();


			// 完了。

			this.tUndoRedo用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tAct小節を挿入する( int n挿入前小節番号 )
		{
			// 挿入する新しい小節の小節長は、直前の（挿入前小節番号-1 の小節）と同じサイズとする。

			double db小節長倍率 = ( n挿入前小節番号 > 0 ) ? this.譜面.Score.t小節長倍率を取得する( n挿入前小節番号 - 1 ) : 1.0;


			// 移動開始。

			this.UndoRedo.tトランザクション記録を開始する();

			#region [ 後方のチップを移動する。]
			//-----------------
			int n挿入に伴う増加量grid = (int) ( this.nGRID_PER_PART * db小節長倍率 );

			foreach( var chip in this.譜面.Score.listチップ )
			{
				if( chip.n小節番号 >= n挿入前小節番号 )
				{
					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 挿入に伴う増加量grid, 任意2 ) => {	// Undo
							変更対象.n小節番号--;
							変更対象.n譜面内絶対位置grid -= (int) 挿入に伴う増加量grid;
						},
						( 変更対象, 変更前, 変更後, 挿入に伴う増加量grid, 任意2 ) => {	// Redo
							変更対象.n小節番号++;
							変更対象.n譜面内絶対位置grid += (int) 挿入に伴う増加量grid;
						},
						chip, null, null, n挿入に伴う増加量grid, null );

					this.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}
			//-----------------
			#endregion

			#region [ 後方の小節長倍率を移動する。]
			//-----------------
			var cc = new CUndoRedoセル<double>(
				null,
				( 変更対象, 変更前, 変更後, 挿入前小節番号, 任意2 ) => {	// Undo
					this.譜面.Score.list小節長倍率.RemoveAt( (int) 挿入前小節番号 );
					this.b未保存 = true;
				},
				( 変更対象, 変更前, 変更後, 挿入前小節番号, 任意2 ) => {	// Redo
					this.譜面.Score.list小節長倍率.Insert( (int) 挿入前小節番号, db小節長倍率 );
					this.b未保存 = true;
				},
				0.0, 0.0, db小節長倍率, n挿入前小節番号, null );
			
			this.UndoRedo.tセルを追加する( cc );
			cc.tRedoを実行する();
			//-----------------
			#endregion


			// 完了。

			this.UndoRedo.tトランザクション記録を終了する();
			this.tUndoRedo用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
			this.b未保存 = true;
		}
		protected void tAct小節を削除する( int n削除する小節番号 )
		{
			double db削除する小節の小節長倍率 = this.譜面.Score.t小節長倍率を取得する( n削除する小節番号 );


			// 削除開始。

			this.UndoRedo.tトランザクション記録を開始する();


			#region [ 削除される小節内のチップをすべて削除する。]
			//-----------------
			for( int i = this.譜面.Score.listチップ.Count - 1; i >= 0; i-- )
			{
				var chip = this.譜面.Score.listチップ[ i ];

				if( chip.n小節番号 == n削除する小節番号 )
				{
					var chip変更前 = new StrokeStyleT.Cチップ( chip );

					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							変更対象.CopyFrom( 変更前 );
							this.譜面.Score.listチップ.Add( 変更対象 );
							this.譜面.Score.listチップ.Sort();
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.譜面.Score.listチップ.Remove( 変更対象 );
							this.b未保存 = true;
						},
						chip, chip変更前, null, null, null );

					this.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}
			//-----------------
			#endregion

			#region [ 後方のチップを移動する。]
			//-----------------
			int n削除に伴う減少量grid = (int) ( this.nGRID_PER_PART * db削除する小節の小節長倍率 );

			foreach( var chip in this.譜面.Score.listチップ )
			{
				if( chip.n小節番号 > n削除する小節番号 )
				{
					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 削除に伴う減少量grid, 任意2 ) => {	// Undo
							変更対象.n小節番号++;
							変更対象.n譜面内絶対位置grid += (int) 削除に伴う減少量grid;
						},
						( 変更対象, 変更前, 変更後, 削除に伴う減少量grid, 任意2 ) => {	// Redo
							変更対象.n小節番号--;
							変更対象.n譜面内絶対位置grid -= (int) 削除に伴う減少量grid;
						},
						chip, null, null, n削除に伴う減少量grid, null );

					this.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}
			//-----------------
			#endregion

			#region [ 後方の小節長倍率を移動する。]
			//-----------------
			var cc = new CUndoRedoセル<double>(
				null,
				( 変更対象, 変更前, 変更後, 削除する小節番号, 任意2 ) => {	// Undo
					this.譜面.Score.list小節長倍率.Insert( (int) 削除する小節番号, 変更前 );
					this.b未保存 = true;
				},
				( 変更対象, 変更前, 変更後, 削除する小節番号, 任意2 ) => {	// Redo
					this.譜面.Score.list小節長倍率.RemoveAt( (int) 削除する小節番号 );
					this.b未保存 = true;
				},
				0.0, db削除する小節の小節長倍率, 0.0, n削除する小節番号, null );

			this.UndoRedo.tセルを追加する( cc );
			cc.tRedoを実行する();
			//-----------------
			#endregion


			// 完了。

			this.UndoRedo.tトランザクション記録を終了する();
			this.tUndoRedo用GUIのEnabledを設定する();
			this.t譜面をリフレッシュする();
		}
		protected void tAct小節の先頭へ移動する( int n小節番号 )
		{
			#region [ 小節番号をクリッピング ]
			//-----------------
			if( n小節番号 < 0 )
				n小節番号 = 0;
			else if( n小節番号 > this.譜面.Score.n最大小節番号 )
				n小節番号 = this.譜面.Score.n最大小節番号;
			//-----------------
			#endregion

			// 垂直スクロールバーを移動させることで画面を移動させる。

			var bar = this.vScrollBar譜面用垂直スクロールバー;
			bar.Value = ( ( bar.Maximum + 1 ) - bar.LargeChange ) - this.譜面.t小節先頭の譜面内絶対位置gridを返す( n小節番号 );
		}


		// アクション補佐メソッド

		protected void tエディタを初期化する()
		{
			this.str編集中のファイル名 = "";


			// 管理パーツを生成する。

			FDK.C共通.tDisposeする( ref this.譜面 );
			this.譜面 = new C譜面( this );	// 譜面は選択・編集モードよりも先に生成すること。
			this.UndoRedo = new CUndoRedo();
			this.選択モード = new C選択モード( this );
			this.編集モード = new C編集モード( this );
			this.クリップボード = new Cクリップボード( this );


			// GUI の初期値を設定する。

			#region [ 基本情報タブ ]
			//-----------------
			this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBox曲名.Clear();
			this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBox説明.Clear();
			//-----------------
			#endregion
			#region [ Viewer 再生 ]
			//-----------------
			this.tViewer再生関連GUIのEnabledを設定する();
			//-----------------
			#endregion
			#region [ ガイド間隔 ]
			//-----------------
			this.tActガイド間隔を変更する( 16 );
			//-----------------
			#endregion
			#region [ 譜面拡大率 ]
			//-----------------
			this.tAct譜面拡大率を変更する( 1 );
			//-----------------
			#endregion
			#region [ Undo/Redo ]
			//-----------------
			this.t次のプロパティ変更がUndoRedoリストに載るようにする();
			this.tUndoRedo用GUIのEnabledを設定する();
			//-----------------
			#endregion
			#region [ 垂直スクロールバー ]
			//-----------------
			this.vScrollBar譜面用垂直スクロールバー.Minimum = 0;
			this.vScrollBar譜面用垂直スクロールバー.Maximum = this.譜面.n全小節の高さgrid - 1;
			this.vScrollBar譜面用垂直スクロールバー.SmallChange = this.nGRID_PER_PART / 16;
			this.vScrollBar譜面用垂直スクロールバー.LargeChange = this.nGRID_PER_PART;
			this.vScrollBar譜面用垂直スクロールバー.Value = this.vScrollBar譜面用垂直スクロールバー.Maximum - this.vScrollBar譜面用垂直スクロールバー.LargeChange;
			//-----------------
			#endregion


			// 最初は編集モードで始める。

			this.t編集モードに切替えて関連GUIを設定する();
			this.b未保存 = false;
		}
		protected DialogResult t未保存なら保存する()
		{
			#region [ 既に保存済みなら帰還。]
			//-----------------
			if( this.b未保存 == false )
				return DialogResult.OK;
			//-----------------
			#endregion

			#region [「編集中のデータを保存しますか？」ダイアログ表示、回答待ち。]
			//-----------------
			var result =
				MessageBox.Show(
					Properties.Resources.MSG_編集中のデータを保存しますか,
					Properties.Resources.MSG_確認ダイアログのタイトル,
					MessageBoxButtons.YesNoCancel,
					MessageBoxIcon.Question,
					MessageBoxDefaultButton.Button1 );
			//-----------------
			#endregion
			#region [ YES なら上書き保存。]
			//-----------------
			if( result == DialogResult.Yes )
				this.tAct上書き保存();
			//-----------------
			#endregion

			#region [ 画面を再描画してダイアログを消去。]
			//-----------------
			this.Refresh();
			//-----------------
			#endregion

			return result;	// ダイアログの結果を返す。
		}

		protected string tファイル保存ダイアログを開いてファイル名を取得する()
		{
			// ダイアログでファイル名を取得。

			var dialog = new SaveFileDialog() {
				Title = "名前をつけて保存",
				Filter = "SSTFファイル(*.sstf)|*.sstf",
				FilterIndex = 1,
				InitialDirectory = this.str作業フォルダ,
			};
			var result = dialog.ShowDialog( this );


			// 画面を再描画。

			this.Refresh();


			// キャンセルされたら "" を返す。

			if( result != DialogResult.OK )
				return "";


			// ファイルの拡張子を .sstf に変更。

			string fileName = dialog.FileName;
			if( Path.GetExtension( fileName ).Length == 0 )
				fileName = Path.ChangeExtension( fileName, ".sstf" );

			return fileName;
		}
		protected void tファイルを読み込む( string fileName )
		{
			this.tエディタを初期化する();

			#region [「読み込み中です」ポップアップを表示する。]
			//-----------------
			var msg	= new CPopupメッセージ( Properties.Resources.MSG_読み込み中です + Environment.NewLine + Properties.Resources.MSG_しばらくお待ち下さい );
			msg.Owner = this;
			msg.Show();
			msg.Refresh();
			//-----------------
			#endregion

			try
			{
				this.譜面.t曲データファイルを読み込む( fileName );

				#region [ 読み込んだファイルに従って内部状態やGUIを更新する。]
				//-----------------
				string str読み込み時の拡張子 = Path.GetExtension( fileName ).ToLower();

				this.str編集中のファイル名 = Path.ChangeExtension( Path.GetFileName( fileName ), ".sstf" );		// 読み込んだファイルの拡張子を .sstf に変換。
				this.str作業フォルダ = Path.GetDirectoryName( fileName );


				// 読み込んだファイルを [ファイル]メニューの最近使ったファイル一覧に追加する。

				this.Config.tファイルを最近使ったファイルの一覧に追加する( Path.Combine( this.str作業フォルダ, this.str編集中のファイル名 ) );
				this.tConfigのRecentUsedFilesをFileメニューへ追加する();


				// 基本情報タブ

				譜面.Score.str背景動画ファイル名 = FDK.C共通.t指定した拡張子を持つファイルを検索し最初に見つけたファイルの絶対パスを返す( this.str作業フォルダ, StrokeStyleT.Config.listデフォルト拡張子 );
				this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBox曲名.Text = 譜面.Score.Header.str曲名;
				this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBox説明.Text = 譜面.Score.Header.str説明文;
				this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBox背景動画.Text = Path.GetFileName( 譜面.Score.str背景動画ファイル名 );
				this.t次のプロパティ変更がUndoRedoリストに載らないようにする(); this.textBoxメモ.Text = ( this.譜面.Score.dicメモ.ContainsKey( 0 ) ) ? this.譜面.Score.dicメモ[ 0 ] : "";


				// ウィンドウのタイトルバーの表示変更（str編集中のファイル名 が確定した後に）

				this.b未保存 = true;		// 以前の状態によらず、確実に更新するようにする。

				if( str読み込み時の拡張子.Equals(".sstf"))		// sstf 以外を読み込んだなら、未保存状態のままとする。
					this.b未保存 = false;
				//-----------------
				#endregion
			}
			catch( InvalidDataException )
			{
				MessageBox.Show(
					Properties.Resources.MSG_対応していないファイルです,
					Properties.Resources.MSG_確認ダイアログのタイトル,
					MessageBoxButtons.OK );
			}

			#region [「読み込み中です」ポップアップを閉じる。 ]
			//-----------------
			msg.Close();
			//-----------------
			#endregion

			// 最後に、ダイアログのゴミなどを消すために再描画。

			this.Refresh();
		}
		
		protected enum Eタブ種別 : int { 基本情報 = 0 }
		protected void tタブを選択する( Eタブ種別 eタブ種別 )
		{
			this.tabControl情報タブコンテナ.SelectedIndex = (int) eタブ種別;
		}

		protected void t次のプロパティ変更がUndoRedoリストに載らないようにする()
		{
			CUndoRedo.bUndoRedoした直後 = true;
		}
		protected void t次のプロパティ変更がUndoRedoリストに載るようにする()
		{
			CUndoRedo.bUndoRedoした直後 = false;
		}
	
		protected void t垂直スクロールバーと譜面の上下位置を調整する()
		{
			var bar = this.vScrollBar譜面用垂直スクロールバー;
			var box = this.pictureBox譜面パネル;
			var panel2 = this.splitContainer分割パネルコンテナ.Panel2;
		

			// 譜面パネルの長さをパネルに合わせて調整する。

			box.ClientSize = new Size( box.ClientSize.Width, panel2.ClientSize.Height - box.Location.Y );


			// 現在のバーの位置を割合で記憶する。
			
			var dbBar率 = (double) bar.Value / (double) ( bar.Maximum - bar.Minimum );


			// 新しい値域を設定した後、バーの位置を記憶しておいた割合で設定。

			bar.Minimum = 0;
			bar.Maximum = this.譜面.n全小節の高さgrid - 1;
			bar.Value = (int) ( ( bar.Maximum - bar.Minimum ) * dbBar率 );
		
			
			// 譜面長さが画面長さより短いならスクロールバーは表示OFF。

			bar.Enabled = bar.Maximum > bar.LargeChange;
		}

		protected void tViewer再生関連GUIのEnabledを設定する()
		{
			if( File.Exists( this.strプレイヤーの絶対パスファイル名 ) )
			{
				// StrokeStyleT.exe (Viewer) が存在するなら、各GUIは有効。

				this.toolStripButton先頭から再生.Enabled = true;
				this.toolStripButton現在位置から再生.Enabled = true;
				this.toolStripButton現在位置からBGMのみ再生.Enabled = true;
				this.toolStripButton再生停止.Enabled = true;

				this.toolStripMenuItem先頭から再生.Enabled = true;
				this.toolStripMenuItem現在位置から再生.Enabled = true;
				this.toolStripMenuItem現在位置からBGMのみ再生.Enabled = true;
				this.toolStripMenuItem再生停止.Enabled = true;
			}
			else
			{
				// StrokeStyleT.exe (Viewer) が存在しないなら、各GUIは無効。

				this.toolStripButton先頭から再生.Enabled = false;
				this.toolStripButton現在位置から再生.Enabled = false;
				this.toolStripButton現在位置からBGMのみ再生.Enabled = false;
				this.toolStripButton再生停止.Enabled = false;

				this.toolStripMenuItem先頭から再生.Enabled = false;
				this.toolStripMenuItem現在位置から再生.Enabled = false;
				this.toolStripMenuItem現在位置からBGMのみ再生.Enabled = false;
				this.toolStripMenuItem再生停止.Enabled = false;
			}
		}
		protected void tConfigのRecentUsedFilesをFileメニューへ追加する()
		{
			#region [ Fileメニューから、「最近使ったファイルの一覧」をクリアする。]
			//-----------------
			for( int i = 0; i < this.toolStripMenuItemファイル.DropDownItems.Count; i++ )
			{
				var item = this.toolStripMenuItemファイル.DropDownItems[ i ];

				// ↓削除したくないサブメニューの一覧。これ以外のサブメニュー項目はすべて削除する。
				if( item != this.toolStripMenuItem新規作成 &&
					item != this.toolStripMenuItem開く &&
					item != this.toolStripMenuItem上書き保存 &&
					item != this.toolStripMenuItem名前を付けて保存 &&
					item != this.toolStripSeparator1 &&
					item != this.toolStripMenuItem終了 )
				{
					this.toolStripMenuItemファイル.DropDownItems.Remove( item );
					i = -1;	// 要素数が変わったので列挙しなおし。RemoveAll() があればなあ...
				}
			}
			//-----------------
			#endregion
			#region [ 表示しないオプション設定であるか、履歴が０件ならここで終了する。]
			//-----------------
			if( !this.Config.ShowRecentUsedFiles || this.Config.RecentUsedFiles.Count == 0 )
				return;
			//-----------------
			#endregion

			#region [ Config が持つ履歴にそって、[ファイル] メニューにサブメニュー項目リストを追加する（ただし最大表示数まで）。 ]
			//-----------------

			// [File] のサブメニューリストに項目が１つでもある場合は、履歴サブメニュー項目の追加の前に「終了」の下にセパレータを入れる。手動で。

			bool bセパレータの追加がまだ = true;


			// すべての「最近使ったファイル」について...

			for( int i = 0; i < this.Config.RecentUsedFiles.Count; i++ )
			{
				#region [ 最大表示数を越えたらここで終了。 ]
				//-----------------
				if( i >= this.Config.MaxOfUsedRecentFiles )
					return;
				//-----------------
				#endregion

				#region [ ファイル名を、サブメニュー項目として [File] メニューに追加する。 ]
				//-----------------
				string path = this.Config.RecentUsedFiles[ i ];

				if( string.IsNullOrEmpty( path ) )
					continue;

				#region [ セパレータの追加がまだなら追加する。]
				//-----------------
				if( bセパレータの追加がまだ )
				{
					this.toolStripMenuItemファイル.DropDownItems.Add( new ToolStripSeparator() { Size = this.toolStripSeparator1.Size } );
					bセパレータの追加がまだ = false;
				}
				//-----------------
				#endregion

				#region [ ToolStripMenuItem を手動で作って [File] のサブメニューリストに追加する。]
				//-----------------
				var item = new ToolStripMenuItem() {
					Name = "最近使ったファイル" + i,
					Size = this.toolStripMenuItem終了.Size,
					Text = "&" + i + " " + path,
				};
				item.Click += new EventHandler( this.toolStripMenuItem最近使ったファイル_Click );
				this.toolStripMenuItemファイル.DropDownItems.Add( item );
				//-----------------
				#endregion

				#region [ 追加したファイルが既に存在していないなら項目を無効化（グレー表示）する。]
				//-----------------
				if( !File.Exists( path ) )
					item.Enabled = false;
				//-----------------
				#endregion

				//-----------------
				#endregion
			}
			//-----------------
			#endregion
		}
		protected Point pt現在のマウス位置を譜面パネル内座標pxに変換して返す()
		{
			Point p = new Point( Cursor.Position.X, Cursor.Position.Y );
			return this.pictureBox譜面パネル.PointToClient( p );
		}

		protected void t現在のチップ音量をツールバーに表示する()
		{
			if( this.dic音量ラベル.ContainsKey( this.n現在のチップ音量 ) )
				this.toolStripLabel音量.Text = this.dic音量ラベル[ this.n現在のチップ音量 ];
			else
				this.toolStripLabel音量.Text = "???";
		}

		
		// イベントメソッド

		#region [ GUIイベント：メインフォーム ]
		//-----------------
		protected void Cメインフォーム_DragDrop( object sender, DragEventArgs e )
		{
			string[] data = (string[]) e.Data.GetData( DataFormats.FileDrop );

			if( data.Length >= 1 )
				this.tAct指定されたファイルを開く( data[ 0 ] );			// Dropされたファイルが複数個あっても、先頭のファイルだけを有効とする。
		}
		protected void Cメインフォーム_DragEnter( object sender, DragEventArgs e )
		{
			if( e.Data.GetDataPresent( DataFormats.FileDrop ) )
			{
				e.Effect = DragDropEffects.Copy;	// ファイルならコピーと見なす（カーソルがコピー型になる）
			}
			else
			{
				e.Effect = DragDropEffects.None;	// ファイルじゃないなら無視（カーソル変化なし）
			}
		}
		protected void Cメインフォーム_FormClosing( object sender, FormClosingEventArgs e )
		{
			if( this.t未保存なら保存する() == DialogResult.Cancel )
			{
				e.Cancel = true;
			}
			else
			{
				this.tActアプリの終了();
			}
		}
		//-----------------
		#endregion
		#region [ GUIイベント：メニューバー ]
		//-----------------

		// File
		protected void toolStripMenuItem新規作成_Click( object sender, EventArgs e )
		{
			this.tAct新規作成();
		}
		protected void toolStripMenuItem開く_Click( object sender, EventArgs e )
		{
			this.tAct開く();
		}
		protected void toolStripMenuItem上書き保存_Click( object sender, EventArgs e )
		{
			this.tAct上書き保存();
		}
		protected void toolStripMenuItem名前を付けて保存_Click( object sender, EventArgs e )
		{
			this.tAct名前を付けて保存();
		}
		protected void toolStripMenuItem終了_Click( object sender, EventArgs e )
		{
			this.tAct終了();
		}
		protected void toolStripMenuItem最近使ったファイル_Click( object sender, EventArgs e )
		{
			// ※このイベントハンドラに対応する「toolStripMenuItem最近使ったファイル」というアイテムはデザイナにはないので注意。
			//   「this.t最近使ったファイルをFileメニューへ追加する()」の中で、手動で作って追加したアイテムに対するハンドラである。
			
			this.tAct指定されたファイルを開く( ( (ToolStripMenuItem) sender ).Text.Substring( 3 ) );
		}

		// Edit
		protected void toolStripMenuItem元に戻す_Click( object sender, EventArgs e )
		{
			this.tAct元に戻す();
		}
		protected void toolStripMenuItemやり直す_Click( object sender, EventArgs e )
		{
			this.tActやり直す();
		}
		protected void toolStripMenuItem切り取り_Click( object sender, EventArgs e )
		{
			this.tAct切り取り();
		}
		protected void toolStripMenuItemコピー_Click( object sender, EventArgs e )
		{
			this.tActコピー();
		}
		protected void toolStripMenuItem貼り付け_Click( object sender, EventArgs e )
		{
			// マウスの位置を譜面パネル内座標で取得。

			Point ptマウスの位置 = this.pt現在のマウス位置を譜面パネル内座標pxに変換して返す();


			// (A) マウスが譜面上になかった → 表示領域下辺から貼り付ける。

			if( ( ( ptマウスの位置.X < 0 ) || ( ptマウスの位置.Y < 0 ) ) || ( ( ptマウスの位置.X > this.sz譜面パネル.Width ) || ( ptマウスの位置.Y > this.sz譜面パネル.Height ) ) )
				this.tAct貼り付け( this.譜面.n現在の譜面表示下辺の譜面内絶対位置grid );

			// (B) マウスが譜面上にある → そこから貼り付ける。

			else
				this.tAct貼り付け( this.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( ptマウスの位置.Y ) );
		}
		protected void toolStripMenuItem削除_Click( object sender, EventArgs e )
		{
			this.tAct削除();
		}
		protected void toolStripMenuItemすべて選択_Click( object sender, EventArgs e )
		{
			this.tActすべて選択();
		}
		protected void toolStripMenuItem選択モード_Click( object sender, EventArgs e )
		{
			this.t選択モードに切替えて関連GUIを設定する();
		}
		protected void toolStripMenuItem編集モード_Click( object sender, EventArgs e )
		{
			this.t編集モードに切替えて関連GUIを設定する();
		}
		protected void toolStripMenuItemモード切替え_Click( object sender, EventArgs e )
		{
			if( this.b選択モードである )
			{
				this.t編集モードに切替えて関連GUIを設定する();
			}
			else
			{
				this.t選択モードに切替えて関連GUIを設定する();
			}
		}
		protected void toolStripMenuItem検索_Click( object sender, EventArgs e )
		{
			this.tAct検索();
		}

		// View
		protected void toolStripMenuItemガイド間隔4分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 4 );
		}
		protected void toolStripMenuItemガイド間隔6分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 6 );
		}
		protected void toolStripMenuItemガイド間隔8分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 8 );
		}
		protected void toolStripMenuItemガイド間隔12分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 12 );
		}
		protected void toolStripMenuItemガイド間隔16分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 16 );
		}
		protected void toolStripMenuItemガイド間隔24分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 24 );
		}
		protected void toolStripMenuItemガイド間隔32分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 32 );
		}
		protected void toolStripMenuItemガイド間隔48分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 48 );
		}
		protected void toolStripMenuItemガイド間隔64分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 64 );
		}
		protected void toolStripMenuItemガイド間隔128分_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 128 );
		}
		protected void toolStripMenuItemガイド間隔フリー_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を変更する( 0 );
		}
		protected void toolStripMenuItemガイド間隔拡大_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を拡大する();
		}
		protected void toolStripMenuItemガイド間隔縮小_Click( object sender, EventArgs e )
		{
			this.tActガイド間隔を縮小する();
		}

		// Play
		protected void toolStripMenuItem先頭から再生_Click( object sender, EventArgs e )
		{
			this.tAct最初から再生する();
		}
		protected void toolStripMenuItem現在位置から再生_Click( object sender, EventArgs e )
		{
			this.tAct現在位置から再生する();
		}
		protected void toolStripMenuItem現在位置からBGMのみ再生_Click( object sender, EventArgs e )
		{
			this.tAct現在位置からBGMのみ再生する();
		}
		protected void toolStripMenuItem再生停止_Click( object sender, EventArgs e )
		{
			this.tAct再生を停止する();
		}

		// Tool
		protected void toolStripMenuItemオプション_Click( object sender, EventArgs e )
		{
			this.tActオプションを設定する();
		}

		// Help
		protected void toolStripMenuItemバージョン_Click( object sender, EventArgs e )
		{
			this.tActバージョンを表示する();
		}

		//-----------------
		#endregion
		#region [ GUIイベント：ツールバー ]
		//-----------------
		protected void toolStripButton新規作成_Click( object sender, EventArgs e )
		{
			this.tAct新規作成();
		}
		protected void toolStripButton開く_Click( object sender, EventArgs e )
		{
			this.tAct開く();
		}
		protected void toolStripButton上書き保存_Click( object sender, EventArgs e )
		{
			this.tAct上書き保存();
		}
		protected void toolStripButton切り取り_Click( object sender, EventArgs e )
		{
			this.tAct切り取り();
		}
		protected void toolStripButtonコピー_Click( object sender, EventArgs e )
		{
			this.tActコピー();
		}
		protected void toolStripButton貼り付け_Click( object sender, EventArgs e )
		{
			// マウスの位置を譜面パネル内座標で取得。

			Point ptマウスの位置 = this.pt現在のマウス位置を譜面パネル内座標pxに変換して返す();


			// (A) マウスが譜面上になかった → 表示領域下辺から貼り付ける。

			if( ( ( ptマウスの位置.X < 0 ) || ( ptマウスの位置.Y < 0 ) ) || ( ( ptマウスの位置.X > this.sz譜面パネル.Width ) || ( ptマウスの位置.Y > this.sz譜面パネル.Height ) ) )
				this.tAct貼り付け( this.譜面.n現在の譜面表示下辺の譜面内絶対位置grid );

			// (B) マウスが譜面上にある → そこから貼り付ける。

			else
				this.tAct貼り付け( this.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( ptマウスの位置.Y ) );
		}
		protected void toolStripButton削除_Click( object sender, EventArgs e )
		{
			this.tAct削除();
		}
		
		protected void toolStripButton元に戻す_Click( object sender, EventArgs e )
		{
			this.tAct元に戻す();
		}
		protected void toolStripButtonやり直す_Click( object sender, EventArgs e )
		{
			this.tActやり直す();
		}

		protected void toolStripComboBox譜面拡大率_SelectedIndexChanged( object sender, EventArgs e )
		{
			this.tAct譜面拡大率を変更する( this.toolStripComboBox譜面拡大率.SelectedIndex + 1 );
		}
		protected void toolStripComboBoxガイド間隔_SelectedIndexChanged( object sender, EventArgs e )
		{
			switch( this.toolStripComboBoxガイド間隔.SelectedIndex )
			{
				case 0:
					this.tActガイド間隔を変更する( 4 );
					return;

				case 1:
					this.tActガイド間隔を変更する( 6 );
					return;

				case 2:
					this.tActガイド間隔を変更する( 8 );
					return;

				case 3:
					this.tActガイド間隔を変更する( 12 );
					return;

				case 4:
					this.tActガイド間隔を変更する( 16 );
					return;

				case 5:
					this.tActガイド間隔を変更する( 24 );
					return;

				case 6:
					this.tActガイド間隔を変更する( 32 );
					return;

				case 7:
					this.tActガイド間隔を変更する( 48 );
					return;

				case 8:
					this.tActガイド間隔を変更する( 64 );
					return;

				case 9:
					this.tActガイド間隔を変更する( 128 );
					return;

				case 10:
					this.tActガイド間隔を変更する( 0 );
					return;
			}

		}
		protected void toolStripButton選択モード_Click( object sender, EventArgs e )
		{
			this.tAct選択モードにする();
		}
		protected void toolStripButton編集モード_Click( object sender, EventArgs e )
		{
			this.tAct編集モードにする();
		}

		protected void toolStripButton先頭から再生_Click( object sender, EventArgs e )
		{
			this.tAct最初から再生する();
		}
		protected void toolStripButton現在位置から再生_Click( object sender, EventArgs e )
		{
			this.tAct現在位置から再生する();
		}
		protected void toolStripButton現在位置からBGMのみ再生_Click( object sender, EventArgs e )
		{
			this.tAct現在位置からBGMのみ再生する();
		}
		protected void toolStripButton再生停止_Click( object sender, EventArgs e )
		{
			this.tAct再生を停止する();
		}

		protected void toolStripButton音量Down_Click( object sender, EventArgs e )
		{
			int n音量 = this.n現在のチップ音量 - 1;
			this.n現在のチップ音量 = ( n音量 < Cメインフォーム.n最小音量 ) ? Cメインフォーム.n最小音量 : n音量;

			this.t現在のチップ音量をツールバーに表示する();
		}
		protected void toolStripButton音量UP_Click( object sender, EventArgs e )
		{
			int n音量 = this.n現在のチップ音量 + 1;
			this.n現在のチップ音量 = ( n音量 > Cメインフォーム.n最大音量 ) ? Cメインフォーム.n最大音量 : n音量;

			this.t現在のチップ音量をツールバーに表示する();
		}
		//-----------------
		#endregion
		#region [ GUIイベント：スプリットパネル、譜面パネル、スクロールバー ]
		//-----------------
		protected void pictureBox譜面パネル_MouseClick( object sender, MouseEventArgs e )
		{
			// フォーカスを得る。

			this.pictureBox譜面パネル.Focus();


			// 選択・編集のいずれかへ処理を引き継ぐ。

			if( this.b選択モードである )
			{
				this.選択モード.MouseClick( e );
			}
			else
			{
				this.編集モード.MouseClick( e );
			}
		}
		protected void pictureBox譜面パネル_MouseDown( object sender, MouseEventArgs e )
		{
			if( this.b選択モードである )
				this.選択モード.MouseDown( e );
		}
		protected void pictureBox譜面パネル_MouseEnter( object sender, EventArgs e )
		{
			#region [ オートフォーカスが有効の場合、譜面にマウスが入ったら譜面がフォーカスを得る。 ]
			//-----------------
			if( this.Config.AutoFocus )
				this.pictureBox譜面パネル.Focus();
			//-----------------
			#endregion
		}
		protected void pictureBox譜面パネル_MouseLeave( object sender, EventArgs e )
		{
			if( this.b編集モードである )
				this.編集モード.MouseLeave( e );
		}
		protected void pictureBox譜面パネル_MouseMove( object sender, MouseEventArgs e )
		{
			// 選択・編集のいずれかの管理者へ処理を引き継ぐ。

			if( this.b選択モードである )
			{
				this.選択モード.MouseMove( e );
			}
			else
			{
				this.編集モード.MouseMove( e );
			}
		}
		protected void pictureBox譜面パネル_Paint( object sender, PaintEventArgs e )
		{
			if( !this.b初期化完了 )
				return;		// 初期化が終わってないのに呼び出されることがあるので、その場合は無視。

			#region [ 小節数が変わってたら、スクロールバーの値域を調整する。]
			//-----------------
			int n全譜面の高さgrid = this.譜面.n全小節の高さgrid;

			if( this.vScrollBar譜面用垂直スクロールバー.Maximum != n全譜面の高さgrid - 1 )	// 小節数が変わっている
			{
				// 譜面の高さ(grid)がどれだけ変わったか？

				int n増加分grid = ( n全譜面の高さgrid - 1 ) - this.vScrollBar譜面用垂直スクロールバー.Maximum;


				// スクロールバーを調整。

				#region [ スクロールバーの状態を新しい譜面の高さに合わせる。]
				//-----------------
				{
					int value = this.vScrollBar譜面用垂直スクロールバー.Value;		// 次の式で Maximum が Value より小さくなると例外が発生するので、
					this.vScrollBar譜面用垂直スクロールバー.Value = 0;				// Value のバックアップを取っておいて、ひとまず 0 にする。
					this.vScrollBar譜面用垂直スクロールバー.Maximum = n全譜面の高さgrid - 1;

					int newValue = value + n増加分grid;


					// オーバーフローしないようクリッピングする。

					if( newValue < 0 )
					{
						this.vScrollBar譜面用垂直スクロールバー.Value = 0;
					}
					else if( newValue >= ( this.vScrollBar譜面用垂直スクロールバー.Maximum - this.vScrollBar譜面用垂直スクロールバー.LargeChange ) )
					{
						this.vScrollBar譜面用垂直スクロールバー.Value = this.vScrollBar譜面用垂直スクロールバー.Maximum - this.vScrollBar譜面用垂直スクロールバー.LargeChange;
					}
					else
					{
						this.vScrollBar譜面用垂直スクロールバー.Value = newValue;
					}
				}
				//-----------------
				#endregion


				// 譜面表示下辺の位置を更新。

				#region [ 譜面表示下辺の位置を更新する。 ]
				//-----------------
				this.譜面.n現在の譜面表示下辺の譜面内絶対位置grid =
					( ( this.vScrollBar譜面用垂直スクロールバー.Maximum - this.vScrollBar譜面用垂直スクロールバー.LargeChange ) + 1 ) - this.vScrollBar譜面用垂直スクロールバー.Value;
				//-----------------
				#endregion
			}
			//-----------------
			#endregion

			#region [ 譜面を描画する。]
			//-----------------
			this.譜面.t描画する( e.Graphics, this.pictureBox譜面パネル );
			//-----------------
			#endregion

			#region [ 現在のモードの Paint() を呼び出す。]
			//-----------------
			if( this.b選択モードである )
			{
				this.選択モード.Paint( e );
			}
			else
			{
				this.編集モード.Paint( e );
			}
			//-----------------
			#endregion
		}
		protected void pictureBox譜面パネル_PreviewKeyDown( object sender, PreviewKeyDownEventArgs e )
		{
			if( e.KeyCode == Keys.Prior )
			{
				#region [ PageUp → 移動量に対応する grid だけ垂直つまみを移動させる。あとはこの移動で生じる ChangedValue イベントで処理。]
				//-----------------
				int n移動すべき数grid = -this.nGRID_PER_PART;
				int n新しい位置 = this.vScrollBar譜面用垂直スクロールバー.Value + n移動すべき数grid;
				int n最小値 = this.vScrollBar譜面用垂直スクロールバー.Minimum;
				int n最大値 = ( this.vScrollBar譜面用垂直スクロールバー.Maximum + 1 ) - this.vScrollBar譜面用垂直スクロールバー.LargeChange;

				if( n新しい位置 < n最小値 )
				{
					n新しい位置 = n最小値;
				}
				else if( n新しい位置 > n最大値 )
				{
					n新しい位置 = n最大値;
				}

				this.vScrollBar譜面用垂直スクロールバー.Value = n新しい位置;
				//-----------------
				#endregion
			}
			else if( e.KeyCode == Keys.Next )
			{
				#region [ PageDown → 移動量に対応する grid だけ垂直つまみを移動させる。あとはこの移動で生じる ChangedValue イベントで処理。]
				//-----------------
				int n移動すべき数grid = this.nGRID_PER_PART;
				int n新しい位置 = this.vScrollBar譜面用垂直スクロールバー.Value + n移動すべき数grid;
				int n最小値 = this.vScrollBar譜面用垂直スクロールバー.Minimum;
				int n最大値 = ( this.vScrollBar譜面用垂直スクロールバー.Maximum + 1 ) - this.vScrollBar譜面用垂直スクロールバー.LargeChange;

				if( n新しい位置 < n最小値 )
				{
					n新しい位置 = n最小値;
				}
				else if( n新しい位置 > n最大値 )
				{
					n新しい位置 = n最大値;
				}

				this.vScrollBar譜面用垂直スクロールバー.Value = n新しい位置;
				//-----------------
				#endregion
			}
			else if( this.b編集モードである )
			{
				this.編集モード.PreviewKeyDown( e );
			}
		}
		
		protected void splitContainer分割パネルコンテナ_MouseWheel( object sender, MouseEventArgs e )
		{
			if( !this.b初期化完了 )
				return;		// 初期化が終わってないのに呼び出されることがあるので、その場合は無視。

			#region [ 移動量に対応する grid だけ垂直つまみを移動させる。あとはこの移動で生じる ChangedValue イベントで処理する。]
			//-----------------
			if( e.Delta == 0 )
				return;		// 移動量なし


			// e.Delta は、スクロールバーを下へ動かしたいときに負、上へ動かしたいときに正となる。

			int n移動すべき行数 = ( -e.Delta * SystemInformation.MouseWheelScrollLines ) / 120;


			// １行＝１拍とする。

			int n移動すべき数grid = n移動すべき行数 * ( this.nGRID_PER_PART / 4 );


			// スクロールバーのつまみを移動。

			int n新しい位置 = this.vScrollBar譜面用垂直スクロールバー.Value + n移動すべき数grid;
			int n最小値 = this.vScrollBar譜面用垂直スクロールバー.Minimum;
			int n最大値 = ( this.vScrollBar譜面用垂直スクロールバー.Maximum + 1 ) - this.vScrollBar譜面用垂直スクロールバー.LargeChange;

			if( n新しい位置 < n最小値 )
			{
				n新しい位置 = n最小値;
			}
			else if( n新しい位置 > n最大値 )
			{
				n新しい位置 = n最大値;
			}

			this.vScrollBar譜面用垂直スクロールバー.Value = n新しい位置;
			//-----------------
			#endregion
		}
		protected void splitContainer分割パネルコンテナ_Panel2_SizeChanged( object sender, EventArgs e )
		{
			if( !this.b初期化完了 )
				return;		// 初期化が終わってないのに呼び出されることがあるので、その場合は無視。

			this.t垂直スクロールバーと譜面の上下位置を調整する();
		}
		protected void splitContainer分割パネルコンテナ_Panel2_Paint( object sender, PaintEventArgs e )
		{
			if( !this.b初期化完了 )
				return;		// 初期化が終わってないのに呼び出されることがあるので、その場合は無視。

			var g = e.Graphics;
			var ptメモ領域左上隅 = new PointF() {
				X = this.譜面.nレーンの合計幅px,
				Y = this.pictureBox譜面パネル.Location.Y,
			};

			#region [ 見出し＜小節メモ＞を描画する。]
			//-----------------
			g.DrawString( Properties.Resources.MSG_小節メモ, this.ftメモ, Brushes.White, PointF.Add( ptメモ領域左上隅, new Size( 24, -24 )/*マージン*/ ) );
			//-----------------
			#endregion

			#region [ 小節メモを描画する。]
			//-----------------
			// グリッド値は 上辺＞下辺 なので注意。
			
			int nパネル下辺grid = this.譜面.n現在の譜面表示下辺の譜面内絶対位置grid;
			int nパネル上辺grid = nパネル下辺grid + ( this.pictureBox譜面パネル.ClientSize.Height * this.nGRID_PER_PIXEL );
			int n開始小節番号 = this.譜面.n現在譜面表示下辺に存在している小節番号;

			for( int n小節番号 = n開始小節番号; n小節番号 <= this.譜面.Score.n最大小節番号; n小節番号++ )
			{
				int n小節の下辺grid = this.譜面.t小節先頭の譜面内絶対位置gridを返す( n小節番号 );
				int n小節の上辺grid = n小節の下辺grid + this.譜面.t小節長をグリッドで返す( n小節番号 );

				if( n小節の下辺grid > nパネル上辺grid )
					break;	// 小節が画面上方にはみ出し切ってしまったらそこで終了。

				if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
				{
					string strメモ = this.譜面.Score.dicメモ[ n小節番号 ];

					string[] lines = strメモ.Split( new string[] { Environment.NewLine }, StringSplitOptions.None );
					int n行数 = lines.Length;

					var ptメモ = new PointF() {
						X = ptメモ領域左上隅.X + 4,	// + 4 はマージン
						Y = ptメモ領域左上隅.Y + ( nパネル上辺grid - n小節の下辺grid ) / this.nGRID_PER_PIXEL - ( n行数 * 16 ),		// 9pt = だいたい16px 
					};
					g.DrawString( strメモ, this.ftメモ, Brushes.White, ptメモ );
				}
			}

			//-----------------
			#endregion
		}

		protected void vScrollBar譜面用垂直スクロールバー_ValueChanged( object sender, EventArgs e )
		{
			if( !this.b初期化完了 )
				return;		// 初期化が終わってないのに呼び出されることがあるので、その場合は無視。

			var bar = vScrollBar譜面用垂直スクロールバー;
			

			if( bar.Enabled )
			{
				// 下辺の位置を再計算。

				this.譜面.n現在の譜面表示下辺の譜面内絶対位置grid = ( ( bar.Maximum + 1 ) - bar.LargeChange ) - bar.Value;


				// 編集モードの場合、カーソルのgrid位置を再計算。

				if( this.b編集モードである )
				{
					this.編集モード.MouseMove(
						new MouseEventArgs( MouseButtons.None, 0, this.編集モード.rc現在のチップカーソル領域.X, this.編集モード.rc現在のチップカーソル領域.Y, 0 ) );
				}


				// メモ用小節番号を再計算。

				this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
				this.numericUpDownメモ用小節番号.Value = this.譜面.n現在カレントラインに存在している小節番号;
				this.t次のプロパティ変更がUndoRedoリストに載るようにする();


				// 小節メモを再描画する。

				this.splitContainer分割パネルコンテナ.Panel2.Refresh();
			}
		}
		//-----------------
		#endregion
		#region [ GUIイベント：譜面右メニュー ]
		//-----------------
		protected void toolStripMenuItem選択チップの切り取り_Click( object sender, EventArgs e )
		{
			this.tAct切り取り();
		}
		protected void toolStripMenuItem選択チップのコピー_Click( object sender, EventArgs e )
		{
			this.tActコピー();
		}
		protected void toolStripMenuItem選択チップの貼り付け_Click( object sender, EventArgs e )
		{
			// メニューが開かれたときのマウスの座標を取得。
			// ※メニューは必ずマウス位置を左上にして表示されるとは限らないため、メニューの表示位置からは取得しないこと。

			Point ptマウスの位置 = this.pt選択モードのコンテクストメニューを開いたときのマウスの位置;

			if( this.譜面.t譜面パネル内X座標pxにある編集レーンを返す( ptマウスの位置.X ) == E編集レーン.Unknown )
				return;		// クリックされた場所にレーンがないなら無視。


			// シナリオを実行。

			this.tAct貼り付け( this.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( ptマウスの位置.Y ) );
		}
		protected void toolStripMenuItem選択チップの削除_Click( object sender, EventArgs e )
		{
			this.tAct削除();
		}
		protected void toolStripMenuItemすべてのチップの選択_Click( object sender, EventArgs e )
		{
			// 編集モードなら強制的に選択モードにする。

			if( this.b編集モードである )
				this.t選択モードに切替えて関連GUIを設定する();


			// 全チップを選択。

			this.選択モード.t全チップを選択する();
		}
		protected void toolStripMenuItem小節長変更_Click( object sender, EventArgs e )
		{
			// メニューが開かれたときのマウスの座標を取得。
			// ※メニューは必ずマウス位置を左上にして表示されるとは限らないため、メニューの表示位置からは取得しないこと。

			Point ptマウスの位置 = this.pt選択モードのコンテクストメニューを開いたときのマウスの位置;

			if( this.譜面.t譜面パネル内X座標pxにある編集レーンを返す( ptマウスの位置.X ) == E編集レーン.Unknown )
				return;		// クリックされた場所にレーンがないなら無視。


			// シナリオを実行。

			this.tAct小節長倍率を変更する( this.譜面.t譜面パネル内Y座標pxにおける小節番号を返す( ptマウスの位置.Y ) );
		}
		protected void toolStripMenuItem小節の挿入_Click( object sender, EventArgs e )
		{
			// メニューが開かれたときのマウスの座標を取得。
			// ※メニューは必ずマウス位置を左上にして表示されるとは限らないため、メニューの表示位置からは取得しないこと。

			Point ptマウスの位置 = this.pt選択モードのコンテクストメニューを開いたときのマウスの位置;

			if( this.譜面.t譜面パネル内X座標pxにある編集レーンを返す( ptマウスの位置.X ) == E編集レーン.Unknown )
				return;		// クリックされた場所にレーンがないなら無視。

	
			// シナリオを実行。

			this.tAct小節を挿入する( this.譜面.t譜面パネル内Y座標pxにおける小節番号を返す( ptマウスの位置.Y ) );
		}
		protected void toolStripMenuItem小節の削除_Click( object sender, EventArgs e )
		{
			// メニューが開かれたときのマウスの座標を取得。
			// ※メニューは必ずマウス位置を左上にして表示されるとは限らないため、メニューの表示位置からは取得しないこと。

			Point ptマウスの位置 = this.pt選択モードのコンテクストメニューを開いたときのマウスの位置;

			if( this.譜面.t譜面パネル内X座標pxにある編集レーンを返す( ptマウスの位置.X ) == E編集レーン.Unknown )
				return;		// クリックされた場所にレーンがないなら無視。

			// シナリオを実行。

			this.tAct小節を削除する( this.譜面.t譜面パネル内Y座標pxにおける小節番号を返す( ptマウスの位置.Y ) );
		}
		//-----------------
		#endregion
		#region [ GUIイベント：基本情報 ]
		//-----------------
		protected void textBox曲名_TextChanged( object sender, EventArgs e )
		{
			#region [ この変更が Undo/Redo したことによるものではない場合、UndoRedoセルを追加or修正する。]
			//-----------------
			if( !CUndoRedo.bUndoRedoした直後 )
			{
				// 最新のセルの所有者が自分？

				var cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();

				if( ( cell != null ) && cell.b所有権がある( this.textBox曲名 ) )
				{
					// (Yes) 最新のセルの "変更後の値" を <現在の値> に更新。

					( (CUndoRedoセル<string>) cell ).変更後の値 = this.textBox曲名.Text;
				}
				else
				{
					// (No) セルを新規追加：
					//      "変更前の値" = <以前の値>
					//      "変更後の値" = <現在の値>
					//      "所有者ID" = 対象となるコンポーネントオブジェクト

					var cc = new CUndoRedoセル<string>(
						this.textBox曲名,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBox曲名.Text = 変更前;
							this.textBox曲名.Focus();
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBox曲名.Text = 変更後;
							this.textBox曲名.Focus();
						},
						null, this.textBox曲名_以前の値, this.textBox曲名.Text, null, null );
					
					this.UndoRedo.tセルを追加する( cc );
					

					// Undo ボタンを有効にする。

					this.tUndoRedo用GUIのEnabledを設定する();
				}
			}
			//-----------------
			#endregion

			this.textBox曲名_以前の値 = this.textBox曲名.Text;		// <以前の値> = <現在の値>
			CUndoRedo.bUndoRedoした直後 = false;
			this.b未保存 = true;

			譜面.Score.Header.str曲名 = this.textBox曲名.Text;	// スコアには随時保存
		}
		protected void textBox曲名_Leave( object sender, EventArgs e )
		{
			#region [ 最新 UndoRedoセル の所有権を放棄する。]
			//-----------------
			var cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();

			if( cell != null )
				cell.t所有権の放棄( this.textBox曲名 );
			//-----------------
			#endregion
		}
		protected string textBox曲名_以前の値 = "";

		protected void textBox説明_TextChanged( object sender, EventArgs e )
		{
			#region [ この変更が Undo/Redo したことによるものではない場合、UndoRedoセルを追加or修正する。]
			//-----------------
			if( !CUndoRedo.bUndoRedoした直後 )
			{
				// 最新のセルの所有者が自分？

				CUndoRedoセルBase cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();

				if( ( cell != null ) && cell.b所有権がある( this.textBox説明 ) )
				{
					// (Yes) 最新のセルの "変更後の値" を <現在の値> に更新。

					( (CUndoRedoセル<string>) cell ).変更後の値 = this.textBox説明.Text;
				}
				else
				{
					// (No) セルを新規追加：
					//      "変更前の値" = <以前の値>
					//      "変更後の値" = <現在の値>
					//      "所有者ID" = 対象となるコンポーネントオブジェクト

					var cc = new CUndoRedoセル<string>(
						this.textBox説明,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBox説明.Text = 変更前;
							this.textBox説明.Focus();
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBox説明.Text = 変更後;
							this.textBox説明.Focus();
						},
						null, this.textBox説明_以前の値, this.textBox説明.Text, null, null );

					this.UndoRedo.tセルを追加する( cc );


					// Undo ボタンを有効にする。

					this.tUndoRedo用GUIのEnabledを設定する();
				}
			}
			//-----------------
			#endregion

			this.textBox説明_以前の値 = this.textBox説明.Text;	// <以前の値> = <現在の値>
			CUndoRedo.bUndoRedoした直後 = false;
			this.b未保存 = true;

			// スコアには随時保存。
			譜面.Score.Header.str説明文 = this.textBox説明.Text;
		}
		protected void textBox説明_Leave( object sender, EventArgs e )
		{
			#region [ 最新 UndoRedoセル の所有権を放棄する。]
			//-----------------
			var cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();

			if( cell != null )
				cell.t所有権の放棄( this.textBox説明 );
			//-----------------
			#endregion
		}
		protected string textBox説明_以前の値 = "";

		protected void textBoxメモ_TextChanged( object sender, EventArgs e )
		{
			#region [ この変更が Undo/Redo したことによるものではない場合、UndoRedoセルを追加or修正する。]
			//-----------------
			if( !CUndoRedo.bUndoRedoした直後 )
			{
				// 最新のセルの所有者が自分？

				CUndoRedoセルBase cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();


				if( ( cell != null ) && cell.b所有権がある( this.textBoxメモ ) )
				{
					// (Yes) 最新のセルの "変更後の値" を <現在の値> に更新。

					( (CUndoRedoセル<string>) cell ).変更後の値 = this.textBoxメモ.Text;
				}
				else
				{
					// (No) セルを新規追加：
					//      "変更前の値" = <以前の値>
					//      "変更後の値" = <現在の値>
					//      "所有者ID" = 対象となるコンポーネントオブジェクト

					var cc = new CUndoRedoセル<string>(
						this.textBoxメモ,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.numericUpDownメモ用小節番号.Value = (decimal) 任意1;
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBoxメモ.Text = 変更前;
							this.textBoxメモ.Focus();

							int n小節番号 = (int) ( (decimal) 任意1 );

							#region [ dicメモ の更新 ]
							//-----------------
							if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
							{
								if( string.IsNullOrEmpty( 変更前 ) )
									this.譜面.Score.dicメモ.Remove( n小節番号 );
								else
									this.譜面.Score.dicメモ[ n小節番号 ] = 変更前;
							}
							else
							{
								if( !string.IsNullOrEmpty( 変更前 ) )
									this.譜面.Score.dicメモ.Add( n小節番号, 変更前 );
							}
							//-----------------
							#endregion

							this.tAct小節の先頭へ移動する( n小節番号 );
							this.splitContainer分割パネルコンテナ.Panel2.Refresh();	// 小節メモをリフレッシュ。

						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.tタブを選択する( Eタブ種別.基本情報 );
							this.numericUpDownメモ用小節番号.Value = (decimal) 任意1;
							this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
							this.textBoxメモ.Text = 変更後;
							this.textBoxメモ.Focus();

							int n小節番号 = (int) ( (decimal) 任意1 );

							#region [ dicメモの更新 ]
							//-----------------
							if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
							{
								if( string.IsNullOrEmpty( 変更後 ) )
									this.譜面.Score.dicメモ.Remove( n小節番号 );
								else
									this.譜面.Score.dicメモ[ n小節番号 ] = 変更後;
							}
							else
							{
								if( !string.IsNullOrEmpty( 変更後 ) )
									this.譜面.Score.dicメモ.Add( n小節番号, 変更後 );
							}
							//-----------------
							#endregion

							this.tAct小節の先頭へ移動する( n小節番号 );
							this.splitContainer分割パネルコンテナ.Panel2.Refresh();	// 小節メモをリフレッシュ。
						},
						null, this.textBoxメモ_以前の値, this.textBoxメモ.Text, (object) this.numericUpDownメモ用小節番号.Value, null );

					this.UndoRedo.tセルを追加する( cc );


					// Undo ボタンを有効にする。

					this.tUndoRedo用GUIのEnabledを設定する();
				}
			}
			//-----------------
			#endregion

			this.textBoxメモ_以前の値 = this.textBoxメモ.Text;	// <以前の値> = <現在の値>

			if( !CUndoRedo.bUndoRedoした直後 )
				this.b未保存 = true;

			CUndoRedo.bUndoRedoした直後 = false;

			#region [ 小節番号に対応するメモを dicメモ に登録する。]
			//-----------------
			{
				int n小節番号 = (int) this.numericUpDownメモ用小節番号.Value;

				if( string.IsNullOrEmpty( this.textBoxメモ.Text ) )
				{
					// (A) 空文字列の場合

					if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
						this.譜面.Score.dicメモ.Remove( n小節番号 );		// 存在してたら削除。
					// 存在してなかったら何もしない。
				}
				else
				{
					// (B) その他の場合

					if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
						this.譜面.Score.dicメモ[ n小節番号 ] = this.textBoxメモ.Text;		// 存在してたら更新。
					else
						this.譜面.Score.dicメモ.Add( n小節番号, this.textBoxメモ.Text );	// 存在してなかったら追加。
				}
			}
			//-----------------
			#endregion
			#region [ もし最終小節だったなら、後ろに４つ小節を加える。]
			//-----------------
			{
				int n小節番号 = (int) this.numericUpDownメモ用小節番号.Value;

				if( n小節番号 == this.譜面.Score.n最大小節番号 )
				{
					this.譜面.t最後の小節の後ろに小節を４つ追加する();
				}
			}
			//-----------------
			#endregion

		}
		protected void textBoxメモ_Leave( object sender, EventArgs e )
		{
			#region [ 最新 UndoRedoセル の所有権を放棄する。]
			//-----------------
			var cell = this.UndoRedo.tUndoするセルを取得して返す・見るだけ();

			if( cell != null )
				cell.t所有権の放棄( this.textBoxメモ );
			//-----------------
			#endregion

			this.splitContainer分割パネルコンテナ.Panel2.Refresh();		// 小節メモをリフレッシュ。
		}
		protected string textBoxメモ_以前の値 = "";
	
		protected void numericUpDownメモ用小節番号_ValueChanged( object sender, EventArgs e )
		{
			// 小節番号にあわせて、textBoxメモにメモを表示する。

			int n小節番号 = (int) this.numericUpDownメモ用小節番号.Value;

			this.t次のプロパティ変更がUndoRedoリストに載らないようにする();
			{
				if( this.譜面.Score.dicメモ.ContainsKey( n小節番号 ) )
					this.textBoxメモ.Text = this.譜面.Score.dicメモ[ n小節番号 ];
				else
					this.textBoxメモ.Text = "";
			}
			this.t次のプロパティ変更がUndoRedoリストに載るようにする();
		}
		//-----------------
		#endregion

		#region [ private ]
		//-----------------
		private Font ftメモ = new Font( FontFamily.GenericSansSerif, 9.0f );
		private readonly Dictionary<int, string> dic音量ラベル = new Dictionary<int, string>() {
			{ 1, "1 (Smallest)" },
			{ 2, "2 (Smaller)" },
			{ 3, "3 (Middle)" },
			{ 4, "4 (Normal)" },
		};
		//-----------------
		#endregion

		//-----------------
		#endregion
	}
}
