﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using FDK;

namespace SSTFEditor
{
	class C編集モード
	{
		public Rectangle rc現在のチップカーソル領域
		{
			get;
			protected set;
		}

		public C編集モード( Cメインフォーム form )
		{
			this.Form = form;
			this.rc現在のチップカーソル領域 = new Rectangle( 0, 0, 0, 0 );

			// 現在のチップ種別番号の初期化。

			this.dicレーン別現在のチップ種別番号 = new Dictionary<E編集レーン, int>();
			this.dicレーン別チップ種別バックアップ = new Dictionary<E編集レーン, int>();

			foreach( var obj in Enum.GetValues(typeof(E編集レーン)))
			{
				var e = (E編集レーン) obj;
				this.dicレーン別現在のチップ種別番号[ e ] = 0;
				this.dicレーン別チップ種別バックアップ[ e ] = 0;
			}


			// レーン別チップ種別対応表の初期化。
			// C譜面.dicチップ編集レーン対応表 を元にするので、譜面を先に生成してから編集モードを生成すること。

			this.dicレーン別チップ種別対応表 = new Dictionary<E編集レーン, List<StrokeStyleT.Eチップ>>();

			foreach( var obj in Enum.GetValues( typeof( E編集レーン ) ) )
				this.dicレーン別チップ種別対応表[ (E編集レーン) obj ] = new List<StrokeStyleT.Eチップ>();
			
			foreach( var kvp in this.Form.譜面.dicチップ編集レーン対応表 )
				this.dicレーン別チップ種別対応表[ kvp.Value ].Add( kvp.Key );
		}

		public void MouseClick( MouseEventArgs e )
		{
			this.tチップを配置または削除する( e );
			this.Form.t譜面をリフレッシュする();
		}
		public void MouseLeave( EventArgs e )
		{
			this.rc現在のチップカーソル領域 = new Rectangle( 0, 0, 0, 0 );
			this.Form.t譜面をリフレッシュする();
		}
		public void MouseMove( MouseEventArgs e )
		{
			#region [ チップカーソルの領域を、現在の位置に合わせて更新する。]
			//-----------------
			var rc以前のチップカーソル領域 = this.rc現在のチップカーソル領域;

			this.e現在チップカーソルがある編集レーン = this.Form.譜面.t譜面パネル内X座標pxにある編集レーンを返す( e.X );
			this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid = this.Form.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( e.Y );

			// カーソルが譜面外にあるならここで終了。

			if( this.e現在チップカーソルがある編集レーン == E編集レーン.Unknown ||
				this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid < 0 )
			{
				this.e現在チップカーソルがある編集レーン = E編集レーン.Unknown;
				this.Form.t譜面をリフレッシュする();	// チップを消去する。
				return;
			}

			// 新しい現在の領域を取得。

			this.rc現在のチップカーソル領域 =
				new Rectangle( 
					this.Form.譜面.t編集レーンのX座標pxを返す( this.e現在チップカーソルがある編集レーン ),
					this.Form.譜面.t譜面内絶対位置gridにおける対象領域内のY座標pxを返す( this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid, this.Form.sz譜面パネル )
						- this.Form.譜面.szチップpx.Height,
					this.Form.譜面.szチップpx.Width,
					this.Form.譜面.szチップpx.Height );
			//-----------------
			#endregion

			this.tチップカーソルのあるレーンに合わせて現在のチップ種別を変更する();

			#region [ 領域が変わってたら譜面を再描画する。]
			//-----------------
			if( !this.rc現在のチップカーソル領域.Equals( rc以前のチップカーソル領域 ) )
				this.Form.t譜面をリフレッシュする();
			//-----------------
			#endregion
		}
		public void Paint( PaintEventArgs e )
		{
			this.tチップカーソルを描画する( e.Graphics, this.Form.e現在のチップ種別 );
		}
		public void PreviewKeyDown( PreviewKeyDownEventArgs e )
		{
			if( e.KeyCode == Keys.Space && !e.Shift )		// SPACE
			{
				#region [ 現在のレーン別チップ種別を、登録順にローテーションする。]
				//-----------------
				
				// インデックスを１つ増やす。インデックスが範囲を超えたら 0 に戻す。
				
				int n = this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ] + 1;
				if( n == this.dicレーン別チップ種別対応表[ this.e現在チップカーソルがある編集レーン ].Count )
					n = 0;
				this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ] = n;


				// 画面へ反映する。

				this.tチップカーソルのあるレーンに合わせて現在のチップ種別を変更する();
				//-----------------
				#endregion

				this.b強制種別使用中 = false;
			}
			else if( e.KeyCode == Keys.Space && e.Shift )	// SHIFT + SPACE
			{
				#region [ 現在のレーン別チップ種別を、特定の種別（２番目）とトグル切り替えする。]
				//-----------------
				if( this.dicレーン別チップ種別対応表[ this.e現在チップカーソルがある編集レーン ].Count > 1 )	// 種別を２つ以上持たないレーンは関係ない。
				{
					if( !this.b強制種別使用中 )
					{
						#region [ 特定の種別（２番目）に切替える。]
						//-----------------

						// 現在のインデックスをバックアップしておく。

						this.dicレーン別チップ種別バックアップ[ this.e現在チップカーソルがある編集レーン ] =
							this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ];

						// 現在のインデックスを強制的に「1」（２番目の要素）にする。

						this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ] = 1;

						//-----------------
						#endregion

						this.b強制種別使用中 = true;
					}
					else
					{
						#region [ 元の種別に戻す。]
						//-----------------

						// 現在のインデックスをバックアップ値に戻す。

						this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ] =
							this.dicレーン別チップ種別バックアップ[ this.e現在チップカーソルがある編集レーン ];

						//-----------------
						#endregion

						this.b強制種別使用中 = false;
					}


					// 画面へ反映する。

					this.tチップカーソルのあるレーンに合わせて現在のチップ種別を変更する();
				}
				//-----------------
				#endregion
			}
		}

		#region [ protected ]
		//-----------------
		protected Cメインフォーム Form;
		protected E編集レーン e現在チップカーソルがある編集レーン;
		protected int n現在のチップカーソルの譜面先頭からのガイド単位の位置grid;
		protected bool b強制種別使用中 = false;
		protected Dictionary<E編集レーン, int> dicレーン別チップ種別バックアップ;
		protected Dictionary<E編集レーン, int> dicレーン別現在のチップ種別番号;
		protected Dictionary<E編集レーン, List<StrokeStyleT.Eチップ>> dicレーン別チップ種別対応表;

		protected void tチップカーソルを描画する( Graphics g, StrokeStyleT.Eチップ eチップ )
		{
			#region [ 事前チェック。]
			//-----------------
			if( this.rc現在のチップカーソル領域.Width <= 0 ||
				this.rc現在のチップカーソル領域.Height <= 0 ||
				this.e現在チップカーソルがある編集レーン == E編集レーン.Unknown ||
				eチップ == StrokeStyleT.Eチップ.Unknown ||
				eチップ == StrokeStyleT.Eチップ.小節線 ||
				eチップ == StrokeStyleT.Eチップ.拍線 ||
				eチップ == StrokeStyleT.Eチップ.小節メモ )
			{
				return;		// 描画しない。
			}
			//-----------------
			#endregion

			this.Form.譜面.tチップを指定領域へ描画する( g, eチップ, this.Form.n現在のチップ音量,  this.rc現在のチップカーソル領域, null );
			this.Form.譜面.tチップの太枠を指定領域へ描画する( g, this.rc現在のチップカーソル領域 );
		}
		protected void tチップカーソルのあるレーンに合わせて現在のチップ種別を変更する()
		{
			int nIndex = this.dicレーン別現在のチップ種別番号[ this.e現在チップカーソルがある編集レーン ];
			this.Form.e現在のチップ種別 = this.dicレーン別チップ種別対応表[ this.e現在チップカーソルがある編集レーン ][ nIndex ];
		}
		protected void tチップを配置または削除する( MouseEventArgs eClick )
		{
			if( this.rc現在のチップカーソル領域.Width <= 0 || this.rc現在のチップカーソル領域.Height <= 0 )
				return;

			if( eClick.Button == MouseButtons.Left )
			{
				#region [ チップを配置する。]
				//-----------------
				bool bCTRL押下 = ( Control.ModifierKeys & Keys.Control ) == Keys.Control;

				if( this.e現在チップカーソルがある編集レーン != E編集レーン.BPM )
				{
					#region [ (A) BPM レーン以外の場合 ]
					//-----------------
					string strチップ内文字列 = null;
					if( this.Form.e現在のチップ種別 == StrokeStyleT.Eチップ.China ) strチップ内文字列 = "C N";
					if( this.Form.e現在のチップ種別 == StrokeStyleT.Eチップ.Splash ) strチップ内文字列 = "S P";

					this.Form.譜面.tチップを配置または置換する(
						this.e現在チップカーソルがある編集レーン,
						this.Form.e現在のチップ種別,
						this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid,
						strチップ内文字列, 
						this.Form.n現在のチップ音量, 
						0.0, 
						false );	// 音量変化は未実装
					//-----------------
					#endregion
				}
				else
				{
					#region [ (B) BPM レーンの場合 ]
					//-----------------
					var dialog = new C数値入力ダイアログ(
						(decimal) this.Form.譜面.db譜面内絶対位置gridにおけるBPMを返す( this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid ),
						0.0001M,
						1000M,
						Properties.Resources.MSG_BPM選択ダイアログの説明文 );

					if( dialog.ShowDialog( this.Form ) != DialogResult.OK )
						return;

					double dbBPM = (double) dialog.dc数値;

					this.Form.譜面.tチップを配置または置換する(
						E編集レーン.BPM,
						StrokeStyleT.Eチップ.BPM,
						this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid,
						dbBPM.ToString("###.##"),
						Cメインフォーム.n最大音量,		// BPM チップは常に最大音量枠
						dbBPM,
						false );
					//-----------------
					#endregion
				}
				//-----------------
				#endregion
			}
			else if( eClick.Button == MouseButtons.Right )
			{
				#region [ チップを削除する。]
				//-----------------
				this.Form.譜面.tチップを削除する(
					this.e現在チップカーソルがある編集レーン, 
					this.n現在のチップカーソルの譜面先頭からのガイド単位の位置grid );
				//-----------------
				#endregion
			}
		}
		//-----------------
		#endregion
	}
}
