﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Diagnostics;
using FDK;

namespace SSTFEditor
{
	class C選択モード : IDisposable
	{
		public C選択モード( Cメインフォーム form )
		{
			this.Form = form;
		}

		public void t個別選択解除( StrokeStyleT.Cチップ cc )
		{
			var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
				null,
				( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
					変更対象.b選択が確定している = true;
					this.Form.b未保存 = true;
				},
				( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
					変更対象.b選択が確定している = false;
					this.Form.b未保存 = true;
				},
				cc, null, null, null, null );

			this.Form.UndoRedo.tセルを追加する( cell );
			cell.tRedoを実行する();

			this.Form.tUndoRedo用GUIのEnabledを設定する();
		}
		public void t全チップを選択する()
		{
			this.Form.UndoRedo.tトランザクション記録を開始する();

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				if( !chip.b選択が確定している )		// 選択されていないすべてのチップを選択する。
				{
					var chip変更前 = new StrokeStyleT.Cチップ( chip );
					var chip変更後 = new StrokeStyleT.Cチップ( chip ) {
						bドラッグ操作により選択中 = false,
						b選択が確定している = true,
					};

					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							変更対象.CopyFrom( 変更前 );
							this.Form.b未保存 = true;
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							変更対象.CopyFrom( 変更後 );
							this.Form.b未保存 = true;
						},
						chip, chip変更前, chip変更後, null, null );

					this.Form.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}

			this.Form.UndoRedo.tトランザクション記録を終了する();
			this.Form.tUndoRedo用GUIのEnabledを設定する();
			this.Form.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
			this.Form.t譜面をリフレッシュする();
		}
		public void t全チップの選択を解除する()
		{
			this.Form.UndoRedo.tトランザクション記録を開始する();

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				if( chip.n枠外レーン数 != 0 || chip.n譜面内絶対位置grid < 0 )
				{
					#region [ 譜面範囲外に出たチップがあれば削除する。]
					//-----------------
					var chip変更前 = this.dic移動開始時のチップ状態[ chip ];
					
					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							変更対象.CopyFrom( chip変更前 );
							this.Form.譜面.Score.listチップ.Add( 変更対象 );
							this.Form.譜面.Score.listチップ.Sort();
							this.Form.b未保存 = true;
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							this.Form.譜面.Score.listチップ.Remove( 変更対象 );
							this.Form.b未保存 = true;
						},
						chip, chip変更前, null, null, null );

					this.Form.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
					//-----------------
					#endregion
				}
				else if( chip.bドラッグ操作により選択中 || chip.b選択が確定している )
				{
					#region [ チップの選択を解除する。]
					//-----------------
					var chip変更前 = new StrokeStyleT.Cチップ( chip );

					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Undo
							変更対象.CopyFrom( 変更前 );
						},
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => {	// Redo
							変更対象.b選択が確定している = false;
							変更対象.bドラッグ操作により選択中 = false;
							変更対象.b移動済 = false;
						},
						chip, chip変更前, null, null, null );

					this.Form.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
					//-----------------
					#endregion
				}
			}

			this.Form.UndoRedo.tトランザクション記録を終了する();
			this.Form.tUndoRedo用GUIのEnabledを設定する();
			this.Form.b未保存 = true;
		}
		public void t検索する()
		{
			var dialog = new C検索条件入力ダイアログ();

			if( dialog.ShowDialog( this.Form ) != DialogResult.OK )
				return;

			int n開始小節番号 = dialog.b小節範囲指定CheckBoxがチェックされている ? dialog.n小節範囲開始番号 : 0;
			int n終了小節番号 = dialog.b小節範囲指定CheckBoxがチェックされている ? dialog.n小節範囲終了番号 : this.Form.譜面.Score.n最大小節番号;

			#region [ 開始・終了が省略されているときは、それぞれ 0・最大小節番号 とみなす。]
			//-----------------
			if( n開始小節番号 < 0 )
				n開始小節番号 = 0;

			if( n終了小節番号 < 0 )
				n終了小節番号 = this.Form.譜面.Score.n最大小節番号;
			//-----------------
			#endregion


			this.Form.UndoRedo.tトランザクション記録を開始する();

			int n選択チップ数 = 0;

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				var e編集レーン = this.Form.譜面.dicチップ編集レーン対応表[ chip.eチップ ];
				
				if( e編集レーン == E編集レーン.Unknown )
					continue;	// 編集レーンを持たないチップは無視する。

				if( chip.n小節番号 >= n開始小節番号 && chip.n小節番号 <= n終了小節番号 )
				{
					if( dialog.b指定した編集レーンが選択されている( e編集レーン ) ||
						dialog.b指定したチップ種別が選択されている( chip.eチップ ) )
					{
						// チップを選択する。

						var chip変更前 = new StrokeStyleT.Cチップ( chip );
						var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
							null,
							( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.b選択が確定している = 変更前.b選択が確定している; },	// Undo
							( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.b選択が確定している = true; },	// Redo
							chip, chip変更前, null, null, null );

						this.Form.UndoRedo.tセルを追加する( cell );
						cell.tRedoを実行する();

						n選択チップ数++;
					}
				}
			}

			this.Form.UndoRedo.tトランザクション記録を終了する();
			this.Form.tUndoRedo用GUIのEnabledを設定する();
			this.Form.t譜面をリフレッシュする();

			#region [ チップ数に応じて結果を表示する。]
			//-----------------
			if( n選択チップ数 > 0 )
			{
				this.Form.t選択チップの有無に応じて編集用GUIのEnabledを設定する();

				MessageBox.Show(
					n選択チップ数 + Properties.Resources.MSG_個のチップが選択されました,
					Properties.Resources.MSG_検索結果ダイアログのタイトル,
					MessageBoxButtons.OK, MessageBoxIcon.Asterisk, MessageBoxDefaultButton.Button1 );
			}
			else
			{
				MessageBox.Show(
					Properties.Resources.MSG_該当するチップはありませんでした,
					Properties.Resources.MSG_検索結果ダイアログのタイトル,
					MessageBoxButtons.OK, MessageBoxIcon.Asterisk, MessageBoxDefaultButton.Button1 );
			}
			//-----------------
			#endregion
		}

		public void MouseClick( MouseEventArgs e )
		{
			if( e.Button == MouseButtons.Right )
				this.Form.t選択モードのコンテクストメニューを表示する( e.X, e.Y );
		}
		public void MouseDown( MouseEventArgs e )
		{
			if( e.Button == MouseButtons.Left )
			{
				StrokeStyleT.Cチップ chip = this.Form.譜面.t譜面パネル内座標pxに存在するチップがあれば返す( e.X, e.Y );


				// (A) チップがないか、未選択のチップがあった場合。

				if( chip == null || !chip.b選択が確定している )
					this.t範囲選択開始処理( e );


				// (B) 選択状態のチップがあり、かつ、クリック時に CTRL が押されている場合。

				else if( ( Control.ModifierKeys & Keys.Control ) == Keys.Control )
					this.t個別選択解除( chip );


				// (C) 選択状態のチップがあり、かつ、クリック時に CTRL が押されていない場合。

				else
					this.t移動開始処理( e );
			}
		}
		public void MouseMove( MouseEventArgs e )
		{
			// (A) 左ボタンが押されながら移動している場合 → 継続処理

			if( e.Button == MouseButtons.Left )
			{
				if( this.b範囲選択のためにドラッグ中 )
					this.t範囲選択継続処理( e );

				else if( this.b移動のためにドラッグ中 )
					this.t移動継続処理( e );
			}


			// (B) 左ボタンが押されずに移動した場合 → 終了処理

			else
			{
				if( this.b範囲選択のためにドラッグ中 )
					this.t範囲選択終了処理( e );

				else if( this.b移動のためにドラッグ中 )
					this.t移動終了処理( e );
			}


			// 譜面を再描画。

			this.Form.t譜面をリフレッシュする();
		}
		public void Paint( PaintEventArgs e )
		{
			if( this.b範囲選択のためにドラッグ中 )
				this.t現在の選択範囲を描画する( e.Graphics );
		}

		#region [ IDisposable 実装 ]
		//-----------------
		public void Dispose()
		{
			C共通.tDisposeする( ref this.br選択領域 );
			C共通.tDisposeする( ref this.pen選択領域 );
		}
		//-----------------
		#endregion

		#region [ protected ]
		//-----------------
		protected struct STレーングリッド座標
		{
			public int n編集レーン番号;			// X座標に相当。
			public int n譜面内絶対位置grid;		// Y座標に相当。
		};

		protected Cメインフォーム Form;
		protected SolidBrush br選択領域 = new SolidBrush( Color.FromArgb( 80, 55, 55, 255 ) );
		protected Pen pen選択領域 = new Pen( Color.LightBlue );

		protected void t現在の選択範囲を描画する( Graphics g )
		{
			var rc現在の選択領域px = new Rectangle() {
				X = Math.Min( this.pt現在の範囲選択用ドラッグ開始位置px.X, this.pt現在の範囲選択用ドラッグ終了位置px.X ),
				Y = Math.Min( this.pt現在の範囲選択用ドラッグ開始位置px.Y, this.pt現在の範囲選択用ドラッグ終了位置px.Y ),
				Width = Math.Abs( (int) ( this.pt現在の範囲選択用ドラッグ開始位置px.X - this.pt現在の範囲選択用ドラッグ終了位置px.X ) ),
				Height = Math.Abs( (int) ( this.pt現在の範囲選択用ドラッグ開始位置px.Y - this.pt現在の範囲選択用ドラッグ終了位置px.Y ) ),
			};

			#region [ クリッピング ]
			//-----------------
			if( rc現在の選択領域px.Width < 0 )
			{
				rc現在の選択領域px.X = this.pt現在の移動用ドラッグ開始位置px.X;
				rc現在の選択領域px.Width = this.pt現在の移動用ドラッグ開始位置px.X - rc現在の選択領域px.X;
			}
			if( rc現在の選択領域px.Height < 0 )
			{
				rc現在の選択領域px.Y = this.pt現在の移動用ドラッグ開始位置px.Y;
				rc現在の選択領域px.Height = this.pt現在の移動用ドラッグ開始位置px.Y - rc現在の選択領域px.Y;
			}
			//-----------------
			#endregion

			if( ( rc現在の選択領域px.Width != 0 ) && ( rc現在の選択領域px.Height != 0 ) )
			{
				g.FillRectangle( this.br選択領域, rc現在の選択領域px );
				g.DrawRectangle( Pens.LightBlue, rc現在の選択領域px );
			}
		}
		protected void t譜面パネルの上下端にきたならスクロールする( MouseEventArgs e )
		{
			const int n上端スクロール発動幅px = 70;
			const int n下端スクロール発動幅px = 50;

			if( e.Y <= n上端スクロール発動幅px )
			{
				double db速度係数X = ( Math.Max( ( n上端スクロール発動幅px - e.Y ), 0 ) ) / (double) n上端スクロール発動幅px;
				double db速度係数Y = ( 1.0 - Math.Cos( ( Math.PI / 2.0 ) * db速度係数X ) ) * 2.0 + 1.0;
				int nスクロール量grid =	(int) ( - db速度係数Y * 180.0 );

				this.Form.t譜面を縦スクロールする( nスクロール量grid );

				if( this.b移動のためにドラッグ中 )
					this.pt現在の移動用ドラッグ開始位置px.Y -= nスクロール量grid / this.Form.nGRID_PER_PIXEL;

				if( this.b範囲選択のためにドラッグ中 )
					this.pt現在の範囲選択用ドラッグ開始位置px.Y -= nスクロール量grid / this.Form.nGRID_PER_PIXEL;
			}
			else if( e.Y >= ( this.Form.sz譜面パネル.Height - n下端スクロール発動幅px ) )
			{
				double db速度係数X = ( Math.Max( ( e.Y - ( this.Form.sz譜面パネル.Height - n上端スクロール発動幅px ) ), 0 ) ) / (double) n下端スクロール発動幅px;
				double db速度係数Y = ( 1.0 - Math.Cos( ( Math.PI / 2.0 ) * db速度係数X ) ) * 2.0 + 1.0;
				int nスクロール量grid =	(int) ( db速度係数Y * 180.0 );

				this.Form.t譜面を縦スクロールする( nスクロール量grid );
				
				if( this.b移動のためにドラッグ中 )
					this.pt現在の移動用ドラッグ開始位置px.Y -= nスクロール量grid / this.Form.nGRID_PER_PIXEL;

				if( this.b範囲選択のためにドラッグ中 )
					this.pt現在の範囲選択用ドラッグ開始位置px.Y -= nスクロール量grid / this.Form.nGRID_PER_PIXEL;
			}
		}
		
		#region [ 移動関連 ]
		//-----------------
		protected bool b移動のためにドラッグ中 = false;
		protected Point pt現在の移動用ドラッグ開始位置px = new Point( 0, 0 );
		protected Point pt現在の移動用ドラッグ終了位置px = new Point( 0, 0 );
		protected Dictionary<StrokeStyleT.Cチップ, StrokeStyleT.Cチップ> dic移動開始時のチップ状態 = new Dictionary<StrokeStyleT.Cチップ, StrokeStyleT.Cチップ>();
		protected STレーングリッド座標 st前回のマウス位置LaneGrid = new STレーングリッド座標();

		protected void t移動開始処理( MouseEventArgs e )
		{
			this.b移動のためにドラッグ中 = true;


			// ドラッグ範囲の初期化。

			this.pt現在の移動用ドラッグ開始位置px.X = this.pt現在の移動用ドラッグ終了位置px.X = e.X;
			this.pt現在の移動用ドラッグ開始位置px.Y = this.pt現在の移動用ドラッグ終了位置px.Y = e.Y;


			// マウス位置（lane×grid）の初期化。

			this.st前回のマウス位置LaneGrid = new STレーングリッド座標() {
				n編集レーン番号 = this.Form.譜面.dicレーン番号[ this.Form.譜面.t譜面パネル内X座標pxにある編集レーンを返す( e.X ) ],
				n譜面内絶対位置grid = this.Form.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( e.Y ),
			};


			// 移動対象チップ（現在選択が確定しているチップ）の「移動開始時のチップの全状態」を、ローカルの Dictionary に控えておく。
			// これは t移動終了処理() で使用する。

			this.dic移動開始時のチップ状態.Clear();
			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				if( chip.b選択が確定している )
					this.dic移動開始時のチップ状態.Add( chip, new StrokeStyleT.Cチップ( chip ) );
			}
		}
		protected void t移動継続処理( MouseEventArgs e )
		{
			// ドラッグ終了位置を現在のマウスの位置に更新。

			this.pt現在の移動用ドラッグ終了位置px.X = e.X;
			this.pt現在の移動用ドラッグ終了位置px.Y = e.Y;


			// スクロールチェック。

			this.t譜面パネルの上下端にきたならスクロールする( e );


			// チップの移動。

			#region [ 現在確定中のチップを移動する。]
			//-----------------

			// 現在の位置を算出。

			var st現在のドラッグ終了位置LaneGrid = new STレーングリッド座標() {
				n編集レーン番号 = this.Form.譜面.dicレーン番号[ this.Form.譜面.t譜面パネル内X座標pxにある編集レーンを返す( this.pt現在の移動用ドラッグ終了位置px.X ) ],
				n譜面内絶対位置grid = this.Form.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridをガイド幅単位で返す( this.pt現在の移動用ドラッグ終了位置px.Y ),
			};


			// 前回位置からの移動量を算出。

			var st移動量LaneGrid = new STレーングリッド座標() {
				n編集レーン番号 = st現在のドラッグ終了位置LaneGrid.n編集レーン番号 - this.st前回のマウス位置LaneGrid.n編集レーン番号,
				n譜面内絶対位置grid = st現在のドラッグ終了位置LaneGrid.n譜面内絶対位置grid - this.st前回のマウス位置LaneGrid.n譜面内絶対位置grid,
			};


			// 前回位置から移動していれば、選択されているすべてのチップを移動させる。

			if( st移動量LaneGrid.n編集レーン番号 != 0 || st移動量LaneGrid.n譜面内絶対位置grid != 0 )
			{
				#region [ 全チップの移動済フラグをリセットする。]
				//-----------------
				foreach( var chip in this.Form.譜面.Score.listチップ )
					chip.b移動済 = false;
				//-----------------
				#endregion

				foreach( var chip in this.Form.譜面.Score.listチップ )
				{
					if( chip.b選択が確定している && !chip.b移動済 )
					{
						if( st移動量LaneGrid.n編集レーン番号 != 0 )
						{
							#region [ チップを横に移動する。]
							//-----------------
							int nレーン数 = Enum.GetValues( typeof( E編集レーン ) ).Length;
							int nチップの現在のレーン番号 = this.Form.譜面.dicレーン番号[ this.Form.譜面.dicチップ編集レーン対応表[ chip.eチップ ] ];
							int nチップの移動後のレーン番号;

							#region [ nチップの移動後のレーン番号 を算出。]
							//-----------------
							if( chip.n枠外レーン数 < 0 )
							{
								nチップの移動後のレーン番号 = nチップの現在のレーン番号 + st移動量LaneGrid.n編集レーン番号;
							}
							else if( chip.n枠外レーン数 > 0 )
							{
								nチップの移動後のレーン番号 = ( ( nレーン数 - 1 ) + chip.n枠外レーン数 ) + st移動量LaneGrid.n編集レーン番号;
							}
							else
							{
								nチップの移動後のレーン番号 = nチップの現在のレーン番号;
							}
							//-----------------
							#endregion
							#region [ nチップの移動後のレーン番号 から、チップのチップ種別 or 枠外レーン数 を修正する。]
							//-----------------
							if( nチップの移動後のレーン番号 < 0 )				// 左にはみ出している
							{
								chip.n枠外レーン数 = nチップの移動後のレーン番号;
							}
							else if( nチップの移動後のレーン番号 >= nレーン数 )	// 右にはみ出している
							{
								chip.n枠外レーン数 = nチップの移動後のレーン番号 - ( nレーン数 - 1 );
							}
							else
							{
								var eチップの移動後の編集レーン = this.Form.譜面.dicレーン番号逆引き[ nチップの移動後のレーン番号 ];

								foreach( var kvp in this.Form.譜面.dicチップ編集レーン対応表 )
								{
									if( kvp.Value == eチップの移動後の編集レーン )	// 対応表で最初に見つけた kvp のチップ種別を採択する。
									{
										chip.eチップ = kvp.Key;
										break;
									}
								}
								chip.n枠外レーン数 = 0;
							}
							//-----------------
							#endregion

							chip.b移動済 = true;
							this.Form.b未保存 = true;
							//-----------------
							#endregion
						}

						if( st移動量LaneGrid.n譜面内絶対位置grid != 0 )
						{
							#region [ チップを縦に移動する。]
							//-----------------
							int n移動後の譜面内絶対位置grid = chip.n譜面内絶対位置grid + st移動量LaneGrid.n譜面内絶対位置grid;

							if( n移動後の譜面内絶対位置grid < 0 )
							{
								// 画面外なので何もしない。
							}
							else if( n移動後の譜面内絶対位置grid >= this.Form.譜面.n全小節の高さgrid )
							{
								chip.n譜面内絶対位置grid = n移動後の譜面内絶対位置grid;		// そこに小節はないが、一応描画する。
							}
							else
							{
								chip.n譜面内絶対位置grid = n移動後の譜面内絶対位置grid;
							}

							chip.b移動済 = true;
							this.Form.b未保存 = true;
							//-----------------
							#endregion
						}
					}
				}


				// 前回位置を現在の位置に更新。

				this.st前回のマウス位置LaneGrid = new STレーングリッド座標() {
					n編集レーン番号 = st現在のドラッグ終了位置LaneGrid.n編集レーン番号,
					n譜面内絶対位置grid = st現在のドラッグ終了位置LaneGrid.n譜面内絶対位置grid,
				};
			}
			//-----------------
			#endregion
		}
		protected void t移動終了処理( MouseEventArgs e )
		{
			this.Form.UndoRedo.tトランザクション記録を開始する();

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				// 選択が確定かつ初期位置から移動しているチップのみ対象とする。

				if( chip.b選択が確定している &&
					( chip.n譜面内絶対位置grid != this.dic移動開始時のチップ状態[ chip ].n譜面内絶対位置grid || chip.eチップ != this.dic移動開始時のチップ状態[ chip ].eチップ ) )
				{
					// ここではまだ、譜面範囲外に出ている（＝枠外レーン数がゼロでない）チップの削除は行わない。（その後再び移動される可能性があるため。）
					// これらの削除処理は「t全チップの選択を解除する()」で行う。

					var chip変更前 = this.dic移動開始時のチップ状態[ chip ];

					int n小節先頭位置grid = 0;
					int n小節番号 = this.Form.譜面.t譜面内絶対位置gridにおける小節の番号と小節先頭の位置gridを返す( chip.n譜面内絶対位置grid, out n小節先頭位置grid );

					var chip変更後 = new StrokeStyleT.Cチップ( chip ) {
						n小節番号 = n小節番号,
						n小節解像度 = (int) ( this.Form.nGRID_PER_PART * this.Form.譜面.Score.t小節長倍率を取得する( n小節番号 ) ),
						n小節内位置 = chip.n譜面内絶対位置grid - n小節先頭位置grid,
					};

					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.CopyFrom( 変更前 ); },	// Undo
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.CopyFrom( 変更後 ); },	// Redo
						chip, chip変更前, chip変更後, null, null );

					this.Form.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}

			this.Form.UndoRedo.tトランザクション記録を終了する();
			this.Form.tUndoRedo用GUIのEnabledを設定する();
			this.b移動のためにドラッグ中 = false;
		}
		//-----------------
		#endregion

		#region [ 範囲選択関連 ]
		//-----------------
		protected bool b範囲選択のためにドラッグ中 = false;
		protected Point pt現在の範囲選択用ドラッグ開始位置px = new Point( 0, 0 );
		protected Point pt現在の範囲選択用ドラッグ終了位置px = new Point( 0, 0 );
		
		protected void t範囲選択開始処理( MouseEventArgs e )
		{
			this.b範囲選択のためにドラッグ中 = true;


			// ドラッグ範囲の初期化。

			this.pt現在の範囲選択用ドラッグ開始位置px.X = this.pt現在の範囲選択用ドラッグ終了位置px.X = e.X;
			this.pt現在の範囲選択用ドラッグ開始位置px.Y = this.pt現在の範囲選択用ドラッグ終了位置px.Y = e.Y;


			// CTRL が押されていない場合、いったん全チップの選択を解除する。

			if( ( Control.ModifierKeys & Keys.Control ) != Keys.Control )
				this.t全チップの選択を解除する();


			// 全チップについて、選択・選択解除の取捨選択。

			this.t現在のドラッグ範囲中のチップをすべて選択状態にしそれ以外は選択を解除する();
		}
		protected void t範囲選択継続処理( MouseEventArgs e )
		{
			// クリッピング。

			int x = e.X;
			int y = e.Y;

			if( x < 0 ) x = 0;
			if( x >= this.Form.sz譜面パネル.Width ) x = this.Form.sz譜面パネル.Width;
			if( y < 0 ) y = 0;
			if( y >= this.Form.sz譜面パネル.Height ) y = this.Form.sz譜面パネル.Height;


			// ドラッグ終了位置を現在のマウスの位置に更新。

			this.pt現在の範囲選択用ドラッグ終了位置px.X = x;
			this.pt現在の範囲選択用ドラッグ終了位置px.Y = y;


			// スクロールチェック。

			this.t譜面パネルの上下端にきたならスクロールする( e );


			// チップの選択 or 選択解除。

			this.t現在のドラッグ範囲中のチップをすべて選択状態にしそれ以外は選択を解除する();
		}
		protected void t範囲選択終了処理( MouseEventArgs e )
		{
			this.b範囲選択のためにドラッグ中 = false;
			this.Form.UndoRedo.tトランザクション記録を開始する();


			// ドラック選択範囲に入っているがまだ選択が確定していないチップを選択確定させる。

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				if( chip.bドラッグ操作により選択中 && !chip.b選択が確定している )		// ドラッグ選択範囲内にあって既に選択が確定されているものについては何もしない。
				{
					var chip変更前 = new StrokeStyleT.Cチップ( chip );
					var chip変更後 = new StrokeStyleT.Cチップ( chip ) {
						bドラッグ操作により選択中 = false,
						b選択が確定している = true,
					};
					var cell = new CUndoRedoセル<StrokeStyleT.Cチップ>(
						null,
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.CopyFrom( 変更前 ); },	// Undo
						( 変更対象, 変更前, 変更後, 任意1, 任意2 ) => { 変更対象.CopyFrom( 変更後 ); },	// Redo
						chip, chip変更前, chip変更後, null, null );

					this.Form.UndoRedo.tセルを追加する( cell );
					cell.tRedoを実行する();
				}
			}

			this.Form.UndoRedo.tトランザクション記録を終了する();
			this.Form.tUndoRedo用GUIのEnabledを設定する();
			this.Form.t選択チップの有無に応じて編集用GUIのEnabledを設定する();
		}

		protected void t現在のドラッグ範囲中のチップをすべて選択状態にしそれ以外は選択を解除する()
		{
			// 現在のドラッグ範囲を lane×grid 座標で算出する。
			// Y座標について： px は上→下、grid は下→上に向かって増加するので注意！

			var rc現在のドラッグ範囲px = new Rectangle() {
				X = Math.Min( this.pt現在の範囲選択用ドラッグ開始位置px.X, this.pt現在の範囲選択用ドラッグ終了位置px.X ),
				Y = Math.Min( this.pt現在の範囲選択用ドラッグ開始位置px.Y, this.pt現在の範囲選択用ドラッグ終了位置px.Y ),
				Width = Math.Abs( (int) ( this.pt現在の範囲選択用ドラッグ開始位置px.X - this.pt現在の範囲選択用ドラッグ終了位置px.X ) ),
				Height = Math.Abs( (int) ( this.pt現在の範囲選択用ドラッグ開始位置px.Y - this.pt現在の範囲選択用ドラッグ終了位置px.Y ) ),
			};
			var rc現在のドラッグ範囲LaneGrid = new Rectangle() {
				X = this.Form.譜面.dicレーン番号[ this.Form.譜面.t譜面パネル内X座標pxにある編集レーンを返す( rc現在のドラッグ範囲px.Left ) ],
				Y = this.Form.譜面.t譜面パネル内Y座標pxにおける譜面内絶対位置gridを返す( rc現在のドラッグ範囲px.Bottom ),
			};
			rc現在のドラッグ範囲LaneGrid.Width = Math.Abs( this.Form.譜面.dicレーン番号[ this.Form.譜面.t譜面パネル内X座標pxにある編集レーンを返す( rc現在のドラッグ範囲px.Right ) ] - rc現在のドラッグ範囲LaneGrid.X );
			rc現在のドラッグ範囲LaneGrid.Height = Math.Abs( rc現在のドラッグ範囲px.Height * this.Form.nGRID_PER_PIXEL );


			// すべてのチップについて、現在のドラッグ範囲 内に存在しているチップは「ドラッグ操作により選択中」フラグを立てる。

			foreach( var chip in this.Form.譜面.Score.listチップ )
			{
				int nチップのレーン番号 = this.Form.譜面.dicレーン番号[ this.Form.譜面.dicチップ編集レーン対応表[ chip.eチップ ] ];
				int nチップの厚さgrid = this.Form.譜面.szチップpx.Height * this.Form.nGRID_PER_PIXEL;

				var rcチップLaneGrid = new Rectangle() {
					X = nチップのレーン番号,
					Y = chip.n譜面内絶対位置grid,
					Width = 0,
					Height = nチップの厚さgrid,
				};

				if( rcチップLaneGrid.Right < rc現在のドラッグ範囲LaneGrid.Left || rc現在のドラッグ範囲LaneGrid.Right < rcチップLaneGrid.Left )
				{
					chip.bドラッグ操作により選択中 = false;		// チップは範囲外のレーンにいる
				}
				else if( rcチップLaneGrid.Bottom < rc現在のドラッグ範囲LaneGrid.Top || rc現在のドラッグ範囲LaneGrid.Bottom < rcチップLaneGrid.Top )
				{
					chip.bドラッグ操作により選択中 = false;		// チップは範囲外のグリッドにいる
				}
				else
					chip.bドラッグ操作により選択中 = true;		// チップは範囲内である
			}
		}
		//-----------------
		#endregion

		//-----------------
		#endregion
	}
}
